/***********************************************************
 * $Id: BeanCallbackWrapper.java 114 2009-08-12 20:29:32Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.InvalidClassException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Initializable;
import org.clazzes.remoting.InvocationContext;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.RemoteInvocation;
import org.springframework.beans.factory.BeanInitializationException;

/**
 * This ServerInvocationHandler delegates remote callbacks to a bean by
 * intelligently discovering the service interface.
 * 
 * @author wglas
 */
public class BeanCallbackWrapper implements CallbackWrapper, Initializable {

	private static final Log log = LogFactory.getLog(BeanCallbackWrapper.class);

	private Class<?> serviceInterface;
	private Map<String,Method> serviceMethods;
	
	/* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#isInitialized()
     */
    public boolean isInitialized() {
       
        return this.serviceMethods != null;
    }

    /* (non-Javadoc)
	 * @see org.clazzes.remoting.Initializable#initialize()
	 */
	public void initialize() throws Exception {
				
		if (log.isDebugEnabled())
			log.debug("Using service interface ["+this.serviceInterface+"].");
		
		Method [] methods = this.serviceInterface.getMethods();
		
		this.serviceMethods = new HashMap<String, Method>();
		
		for (Method method : methods) {
			
			Method om = this.serviceMethods.get(method.getName());
			
			if (om != null)
				throw new InvalidClassException("Method ["+om+"] is duplicated in interface ["+this.serviceInterface.getName()+"] (duplicate is ["+method+"]).");
			
			if (log.isDebugEnabled())
				log.debug("Adding method ["+method +"] to the list of service methods.");
				
			this.serviceMethods.put(method.getName(),method);
		}
	}

	private class WrapperInvocationHandler implements InvocationHandler {
		
		final Object target;

		public WrapperInvocationHandler(Object target) {
			super();
			this.target = target;
		}
		
		public Object invoke(InvocationContext ctx, Object arg,
				Map<String, Object> metadata) throws Throwable {
	
			RemoteInvocation ri = (RemoteInvocation) arg;
			
			try {
			    Method m = BeanCallbackWrapper.this.serviceMethods.get(ri.getMethodName());
			    return m.invoke(this.target,ri.getArguments());
			} catch (InvocationTargetException e) {
			    throw e.getCause();
			}
		}
		
	}
	
	/* (non-Javadoc)
	 * @see at.cdes.remoting.client.InterfaceCallbackHandler#getWrapper(java.lang.Object)
	 */
	public InvocationHandler getWrapper(final Object target) {
		
		return new WrapperInvocationHandler(target);
	}
		
	/* (non-Javadoc)
	 * @see at.cdes.remoting.client.InterfaceCallbackHandler#getServiceInterface()
	 */
	public Class<?> getServiceInterface() {
		return this.serviceInterface;
	}

	/**
	 * @param serviceInterface the serviceInterface to set
	 */
	public void setServiceInterface(Class<?> serviceInterface) {
		
		if (!serviceInterface.isInterface())
			throw new BeanInitializationException("Class ["+serviceInterface.getName()+"] is not an interface.");
		
		this.serviceInterface = serviceInterface;
	}

}
