/***********************************************************
 * $Id: BeanImporter.java 67 2009-05-02 09:35:34Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.server.UID;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Initializable;
import org.clazzes.remoting.RemoteInvocation;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;

/**
 * Imports a service exported by {@link BeanInvocationHandler} via remoting.
 * 
 * @author wglas
 */
public class BeanImporter implements Initializable {

	private static final Log log = LogFactory.getLog(BeanImporter.class);
		
	private ClientManager clientManager;
	private Class<?> serviceInterface;
	private String subSystem;
	private Object proxy;
	private List<CallbackWrapper> callbackWrappers;
	private ReturnCallbackProxyFactory returnCallbackProxyFactory;
	
	private static class CallbackArgument {
		
		public final int position;
		public final CallbackWrapper callbackHandler;
		
		public CallbackArgument(int position,
				CallbackWrapper callbackHandler) {
			super();
			this.position = position;
			this.callbackHandler = callbackHandler;
		}
	};
	
	private Map<Method,List<CallbackArgument>> callbackMethods;
	
	private class ClientInvocationHandler implements InvocationHandler {

		/* (non-Javadoc)
		 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
		 */
		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
			
			ClientHolder ch = ClientHolder.getContext();
			
			Client client;
				
			if (ch == null) {
				client = BeanImporter.this.clientManager.activateClient();
			} else {
				client = ch.getClient();
			}
			
			try {
				Object res;
				
				try {
					
					if (log.isDebugEnabled())
						log.debug("Calling method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"]...");
					
					List<CallbackArgument> callbackArguments = null;
					
					if (BeanImporter.this.callbackMethods!=null)
						callbackArguments = BeanImporter.this.callbackMethods.get(method);
					
					if (callbackArguments == null) {
						res = client.invoke(BeanImporter.this.subSystem,new RemoteInvocation(method.getName(),args),null);
					}
					else {
						
						Object[] newArgs  = new Object[args.length];
						
						int ilast = 0;
						
						UID[] cbIds = new UID[callbackArguments.size()];
						
						int nCallbacks = 0;
						
						for (CallbackArgument ca : callbackArguments) {
							
							Object arg = args[ca.position];
							
							if (arg != null) {
								
								if (ca.position > ilast)
									System.arraycopy(args,ilast,newArgs,ilast,ca.position-ilast);
								
								cbIds[nCallbacks] = client.registerCallback(
										ca.callbackHandler.getWrapper(args[ca.position]));
							
								if (log.isDebugEnabled())
									log.debug("Registered callback no. ["+nCallbacks+"] for argument ["+ca.position+"] with UID ["+cbIds[nCallbacks]+"]...");

								newArgs[ca.position] = cbIds[nCallbacks];
				
								++nCallbacks;
								ilast = ca.position + 1;
							}
						}
						
						if (ilast < args.length)
							System.arraycopy(args,ilast,newArgs,ilast,args.length-ilast);
						
						try {
							res = client.invoke(BeanImporter.this.subSystem,new RemoteInvocation(method.getName(),newArgs),null);
							
						} finally {
							
							for (int icb = 0; icb<nCallbacks; ++icb) {
								
								CallbackArgument ca =  callbackArguments.get(icb);
							
								if (log.isDebugEnabled())
									log.debug("deregistered callback for argument ["+ca.position+"] with GUID ["+cbIds[icb]+"]...");
								
								client.deregisterCallback(cbIds[icb]);
							}
						}
					}
					
					if (log.isDebugEnabled())
						log.debug("Successfully called method ["+method.getName()+"] on subsystem ["+BeanImporter.this.subSystem+"].");

				} catch (IOException t) {
					
					// reconnect to the server, if there's an exception from the transport layer.
					if (ch == null) {
						log.error("Disconnecting client upon transport error.",t);
						client.disconnect();
					} else {
						
						log.error("Initiating deferred disconnect of client upon transport error.",t);
						ch.setCloseOnPassivate(true);
					}
					
					
					throw t;
				}
				catch (InvocationTargetException ite)  {
					
					// combine the foreign and the local stack trace.
					Throwable e = ite.getCause();
					
					StackTraceElement[] foreignStack = e.getStackTrace();
					e.fillInStackTrace();
					StackTraceElement[] myStack = e.getStackTrace();
					StackTraceElement[] combinedStack = new StackTraceElement[foreignStack.length+myStack.length];
					System.arraycopy(foreignStack,0,combinedStack,0,foreignStack.length);
					System.arraycopy(myStack,0,combinedStack,foreignStack.length,myStack.length);
					e.setStackTrace(combinedStack);
					throw e;			
				}
				
				if (res != null && BeanImporter.this.returnCallbackProxyFactory != null)
					return BeanImporter.this.returnCallbackProxyFactory.createProxy(method.getReturnType(),res);
				
				return res;
				
			} finally {
				if (ch == null)
					BeanImporter.this.clientManager.passivateClient(client);
			}
		}
	}
	
	/* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#isInitialized()
     */
    public boolean isInitialized() {
        
        return this.proxy != null;
    }

    /* (non-Javadoc)
	 * @see org.clazzes.remoting.Initializable#initialize()
	 */
	public void initialize() throws Exception {
		
		if (this.callbackWrappers != null) {
			
			this.callbackMethods = new HashMap<Method, List<CallbackArgument>>();
		
			for (Method method : this.serviceInterface.getMethods()) {
			
				List<CallbackArgument> callbackArguments = null;
			
				for (int iarg = 0; iarg < method.getParameterTypes().length; ++iarg){
					
					Class<?> arg = method.getParameterTypes()[iarg];
					
					for (CallbackWrapper handler : this.callbackWrappers) {
					
						if (handler.getServiceInterface().isAssignableFrom(arg))
						{
							if (callbackArguments == null)
								callbackArguments = new ArrayList<CallbackArgument>();
							
							if (log.isDebugEnabled())
								log.debug("Argument ["+iarg+"] of method ["+method+"] will be treated as callback with interface ["+handler.getServiceInterface().getName()+"].");
							
							callbackArguments.add(new CallbackArgument(iarg,handler));
						}
					}
				}
				
				if (callbackArguments != null)
					this.callbackMethods.put(method,callbackArguments);
			}
		}
		
		this.proxy=
			Proxy.newProxyInstance(BeanImporter.class.getClassLoader(),
					new Class<?>[] {this.serviceInterface},new ClientInvocationHandler());		
	}

	/**
	 * @return The proxy, which has been set up in {@link #initialize()}.
	 */
	public Object getProxy() {
		
		return this.proxy;
	}

	/**
	 * @return the clientManager
	 */
	public ClientManager getClientManager() {
		return this.clientManager;
	}

	/**
	 * @param clientManager the clientManager to set
	 */
	public void setClientManager(ClientManager clientManager) {
		this.clientManager = clientManager;
	}

	/**
	 * @return the jboss remoting subsystem.
	 * @see Client#setSubsystem(String)
	 */
	public String getSubSystem() {
		return this.subSystem;
	}

	/**
	 * @param subSystem the subSystem to set
	 */
	public void setSubSystem(String subSystem) {
		this.subSystem = subSystem;
	}

	/**
	 * @return the service interface implemented by the proxy.
	 */
	public Class<?> getServiceInterface() {
		return this.serviceInterface;
	}

	/**
	 * @param serviceInterface the serviceInterface to set.
	 */
	public void setServiceInterface(Class<?> serviceInterface) {
		this.serviceInterface = serviceInterface;
	}

	/**
	 * @return the callbackHandlers
	 */
	public List<CallbackWrapper> getCallbackWrappers() {
		return this.callbackWrappers;
	}

	/**
	 * @param callbackHandlers the callbackHandlers to set
	 */
	public void setCallbackWrappers(List<CallbackWrapper> callbackHandlers) {
		this.callbackWrappers = callbackHandlers;
	}

	/**
	 * @return the returnCallbackProxyFactory
	 */
	public ReturnCallbackProxyFactory getReturnCallbackProxyFactory() {
		return this.returnCallbackProxyFactory;
	}

	/**
	 * @param returnCallbackProxyFactory the returnCallbackProxyFactory to set
	 */
	public void setReturnCallbackProxyFactory(
			ReturnCallbackProxyFactory returnCallbackProxyFactory) {
		this.returnCallbackProxyFactory = returnCallbackProxyFactory;
	}
}
