/***********************************************************
 * $Id: OpenClientContextInViewFilter.java 68 2009-05-02 21:23:05Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.TimeoutException;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;
import org.springframework.web.filter.OncePerRequestFilter;

/**
 * Servlet 2.4 Filter that binds a remote Session to the thread for the entire
 * processing of the request.
 *
 * This filter makes a remote Session context available via the current thread,
 * which will be autodetected by {@link BeanImporter}.
 * 
 * In order to use this class you need to use the optional spring dependency.
 *
 * @author wglas
 */
public class OpenClientContextInViewFilter extends OncePerRequestFilter {

    private static Log log = LogFactory.getLog(OpenClientContextInViewFilter.class);

    public static final String DEFAULT_CLIENT_MANAGER_BEAN_NAME = "clientManager";
    public static final String DEFAULT_SESSION_CONTEXT_SUBSYTEM = "sessionContext";

    private String clientManagerBeanName = DEFAULT_CLIENT_MANAGER_BEAN_NAME;
    private String sessionContextSubSystem = DEFAULT_SESSION_CONTEXT_SUBSYTEM;
    private String sessionAuthenticatorBeanName;
    private String overrideFormCharset;

    private String excludeServices;
    private Set<String> excludeServicesSet;
    
    /**
     * Set the bean name of the SessionContextFactory to fetch from Spring's
     * root application context. Default is "sessionContextFactory".
     * @see #DEFAULT_CLIENT_MANAGER_BEAN_NAME
     */
    public void setClientManagerBeanName(String sessionContextFactoryBeanName) {
        this.clientManagerBeanName = sessionContextFactoryBeanName;
    }
    
    /**
     * Return the bean name of the SessionContextFactory to fetch from Spring's
     * root application context.
     */
    protected String getClientManagerBeanName() {
        return this.clientManagerBeanName;
    }
    
    protected ClientHolder makeClientHolder(Client client) {
        return new ClientHolder(client);
    }

    /**
	 * @return the sessionContextSubSystem
	 */
	protected String getSessionContextSubSystem() {
		return this.sessionContextSubSystem;
	}

	/**
	 * @param sessionContextSubSystem the sessionContextSubSystem to set
	 */
	public void setSessionContextSubSystem(String sessionContextSubSystem) {
		this.sessionContextSubSystem = sessionContextSubSystem;
	}
	
	private static String buildRequestLog(HttpServletRequest request) {
		
		if (request.getQueryString() != null)
			return request.getRequestURI() + "?" + request.getQueryString();
		
		return request.getRequestURI();
	}

    private void openSession(SessionAuthenticator sessionAuthenticator, Client client)
    throws InvocationTargetException, IOException {
    	
    	if (sessionAuthenticator == null) {
    		
			client.invoke(this.getSessionContextSubSystem(),SessionCmd.OPEN_SESSION,null);
		}
		else {
			
			client.invoke(this.getSessionContextSubSystem(),SessionCmd.OPEN_SESSION,
					sessionAuthenticator.createCredentials(client.getClientUID()));
		}
    }

    protected void doFilterInternal(
            HttpServletRequest request, HttpServletResponse response, FilterChain filterChain)
    throws ServletException, IOException {

		if (this.excludeServicesSet != null) {
			
			// override the character encoding used to decode POST form-variables.
			if (this.getOverrideFormCharset() != null)
				request.setCharacterEncoding(this.getOverrideFormCharset());
			
			String svc = request.getParameter("service");
			
			if (svc != null && this.excludeServices.contains(svc)) {
			
				if (log.isDebugEnabled())
            		log.debug("Skipping session interception for request ["+buildRequestLog(request)+"].");        	

				filterChain.doFilter(request, response);
				return;
			}
		}
		
        ClientManager clientManager = lookupClientManager();

        Client client;
		try {
			client = clientManager.activateClient();
		} catch (TimeoutException e) {
			throw new ServletException("Unable to get a remoting client.",e);
		}
        
		ClientHolder context = this.makeClientHolder(client);
		
		if (log.isDebugEnabled())
    		log.debug("Binding client session context to current thread ["+Thread.currentThread().getName()+"].");

		ClientHolder.setContext(context);
        
        try {
        	
        	try {
            	if (log.isDebugEnabled())
            		log.debug("Opening remote session for current thread ["+Thread.currentThread().getName()+"] for request ["+buildRequestLog(request)+"].");        	

            	SessionAuthenticator sessionAuthenticator = this.lookupSessionAuthenticator();
            	
            	if (log.isDebugEnabled())
            		log.debug("Opening remote session for current thread ["+Thread.currentThread().getName()+"].");        	

            	try {
            		this.openSession(sessionAuthenticator, client);
            	} catch (IOException e) {
            		log.warn("Transport error while opening session, trying to reconnect.",e);
            		client.disconnect();
            		client.connect();
            		this.openSession(sessionAuthenticator, client);
            		log.info("Successfully reconnected client.");
            	}

            	
			} catch (Throwable e) {
				throw new ServletException("Unable to open a remote session of current thread ["+Thread.currentThread().getName()+"].",e);
			}
        	
            filterChain.doFilter(request, response);
        }
        
        finally {
        	
        	if (client.isConnected()) {
        		try {
        			if (log.isDebugEnabled())
        				log.debug("Closing remote session of current thread ["+Thread.currentThread().getName()+"] for request ["+buildRequestLog(request)+"].");        	
        			
        			client.invoke(this.getSessionContextSubSystem(),SessionCmd.CLOSE_SESSION,null);
        			
        			if (context.isCloseOnPassivate()) {
        				
        				if (log.isDebugEnabled())
            				log.debug("Closing client connection of current thread ["+Thread.currentThread().getName()+"] due to earlier transport errors.");        	
        				client.disconnect();
        			}
        		}
        		catch (Throwable e) {
        			log.warn("Error closing remote seession of current thread ["+Thread.currentThread().getName()+"].",e);
        		}
        	}
        	       	
        	clientManager.passivateClient(client);
        	
        	if (log.isDebugEnabled())
        		log.debug("Unbinding client session context Session from current thread ["+Thread.currentThread().getName()+"].");        	
                   
        	ClientHolder.removeContext();
        }
	}
    
    /**
     * Look up the ClientManager that this filter should use.
     * The default implementation looks for a bean with the specified name
     * in Spring's root application context.
     * @return the ClientManager to use.
     * @see #getClientManagerBeanName()
     */
    protected ClientManager lookupClientManager() {
       
        WebApplicationContext wac =
            WebApplicationContextUtils.getRequiredWebApplicationContext(getServletContext());
        return (ClientManager) wac.getBean(getClientManagerBeanName(), ClientManager.class);
    }

	/**
	 * @return A comma-separated list of tapestry service names to exclude from session interception.
	 */
	public String getExcludeServices() {
		return this.excludeServices;
	}

	/**
	 * @param excludeServices A comma-separated list of tapestry service names to exclude
	 *                        from session interception.
	 */
	public void setExcludeServices(String excludeServices) {
		this.excludeServices = excludeServices;
		
		String[] svcs = this.excludeServices.split(",");
		
		this.excludeServicesSet = new HashSet<String>(svcs.length);
		
		for (String svc : svcs) {
		
			this.excludeServicesSet.add(svc);
		}
	}

	/**
	 * @return The name of the session authenticator bean.
	 */
	public String getSessionAuthenticatorBeanName() {
		return this.sessionAuthenticatorBeanName;
	}

	/**
	 * @param sessionAuthenticatorBeanName The name of the session authenticator bean.
	 */
	public void setSessionAuthenticatorBeanName(String sessionAuthenticatorBeanName) {
		this.sessionAuthenticatorBeanName = sessionAuthenticatorBeanName;
	}
	
	/**
     * Look up the SessionAuthenticator that this filter should use.
     * The default implementation looks for a bean with the specified name
     * in Spring's root application context.
     * @return the SessionAuthenticator to use or <code>null</code>, if the client authenticator
     *         bean name has not been configured.
     * @see #getSessionAuthenticatorBeanName()
     */
    protected SessionAuthenticator lookupSessionAuthenticator() {
       
    	if (this.getSessionAuthenticatorBeanName() == null)
    		return null;
    	
        WebApplicationContext wac =
            WebApplicationContextUtils.getRequiredWebApplicationContext(getServletContext());
        return (SessionAuthenticator)wac.getBean(this.getSessionAuthenticatorBeanName(), SessionAuthenticator.class);
    }

	/**
	 * @return The charset used to decode POST variables. This is needed, if
	 *         {@link #getExcludeServices()} is not null and you experience encoding
	 *         problem in the downstream web application. The configured string is passed to
	 *         {@link ServletRequest#setCharacterEncoding(String)}.
	 */
	public String getOverrideFormCharset() {
		return this.overrideFormCharset;
	}

	/**
	 * @param overrideFormCharset The charset used to decode POST variables. This is needed, if
	 *         {@link #getExcludeServices()} is not null and you experience encoding
	 *         problem in the downstream web application. The configured string is passed to
	 *         {@link ServletRequest#setCharacterEncoding(String)}.
	 */
	public void setOverrideFormCharset(String overrideFormCharset) {
		this.overrideFormCharset = overrideFormCharset;
	}
}
