/***********************************************************
 * $Id: OpenClientSessionInterceptor.java 68 2009-05-02 21:23:05Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.client.Client;
import org.clazzes.remoting.client.ClientManager;

/**
 * This interceptor bind a remote session and a client to the current thread upon
 * the end of a method invocation.
 *
 * In order to use this class you need to use the optional spring dependency or
 * use aopalliance-1.0
 * 
 * @author wglas
 */
public class OpenClientSessionInterceptor implements MethodInterceptor {

    private static Log log = LogFactory.getLog(OpenClientSessionInterceptor.class);

    public static final String DEFAULT_SESSION_CONTEXT_SUBSYTEM = "sessionContext";

    private String sessionContextSubSystem = DEFAULT_SESSION_CONTEXT_SUBSYTEM;

    private ClientManager clientManager;
    private SessionAuthenticator sessionAuthenticator;

    protected ClientHolder makeClientHolder(Client client) {
        return new ClientHolder(client);
    }
    
    private void openSession(Client client) throws InvocationTargetException, IOException {
    	
    	if (this.sessionAuthenticator == null) {
    		
			client.invoke(this.getSessionContextSubSystem(),SessionCmd.OPEN_SESSION,null);
		}
		else {
			
			client.invoke(this.getSessionContextSubSystem(),SessionCmd.OPEN_SESSION,
					this.sessionAuthenticator.createCredentials(client.getClientUID()));
		}
    }
    
    /* (non-Javadoc)
     * @see org.aopalliance.intercept.MethodInterceptor#invoke(org.aopalliance.intercept.MethodInvocation)
     */
    public Object invoke(MethodInvocation invocation) throws Throwable {
       
        Client client;
        client = this.clientManager.activateClient();
        
		ClientHolder context = this.makeClientHolder(client);
	    
		if (log.isDebugEnabled())
    		log.debug("Binding client session context to current thread ["+Thread.currentThread().getName()+"].");

		ClientHolder.setContext(context);

		try {
        	
        	if (log.isDebugEnabled())
        		log.debug("Opening remote session for current thread ["+Thread.currentThread().getName()+"].");        	

        	try {
        		this.openSession(client);
        	} catch (IOException e) {
        		log.warn("Transport error while opening session, trying to reconnect.",e);
        		client.disconnect();
        		client.connect();
        		this.openSession(client);
        		log.info("Successfully reconnected client.");
        	}
        	
            Object ret = invocation.proceed();
            return ret;
        }
        finally {
        	
        	if (client.isConnected()) {
        		try {
        			if (log.isDebugEnabled())
        				log.debug("Closing remote session of current thread ["+Thread.currentThread().getName()+"].");        	

        			client.invoke(this.getSessionContextSubSystem(),SessionCmd.CLOSE_SESSION,null);
        			
        			if (context.isCloseOnPassivate()) {
        				
        				if (log.isDebugEnabled())
            				log.debug("Closing client connection of current thread ["+Thread.currentThread().getName()+"] due to earlier transport errors.");        	
        				client.disconnect();
        			}
        		}
        		catch (Throwable e) {
        			log.warn("Error closing remote session of current thread ["+Thread.currentThread().getName()+"].",e);
        		}
        	}
        	       
        	this.clientManager.passivateClient(client);
        	
        	if (log.isDebugEnabled())
        		log.debug("Unbinding client session context Session from current thread ["+Thread.currentThread().getName()+"].");        	
                   
        	ClientHolder.removeContext();
        }
    }

	/**
	 * @return the clientManager
	 */
	public ClientManager getClientManager() {
		return this.clientManager;
	}

	/**
	 * @param clientManager the clientManager to set
	 */
	public void setClientManager(ClientManager clientManager) {
		this.clientManager = clientManager;
	}

	/**
	 * @return the sessionContextSubSystem
	 */
	public String getSessionContextSubSystem() {
		return this.sessionContextSubSystem;
	}

	/**
	 * @param sessionContextSubSystem the sessionContextSubSystem to set
	 */
	public void setSessionContextSubSystem(String sessionContextSubSystem) {
		this.sessionContextSubSystem = sessionContextSubSystem;
	}

	/**
	 * @return The authenticator used to generate authentication metadat for the server.
	 */
	public SessionAuthenticator getSessionAuthenticator() {
		return this.sessionAuthenticator;
	}

	/**
	 * @param sessionAuthenticator The session authenticator to set.
	 */
	public void setSessionAuthenticator(SessionAuthenticator sessionAuthenticator) {
		this.sessionAuthenticator = sessionAuthenticator;
	}
}
