/***********************************************************
 * $Id: ReturnCallbackInvocationHandler.java 68 2009-05-02 21:23:05Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 29.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.beans;

import java.io.InvalidClassException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.server.UID;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Destroyable;
import org.clazzes.remoting.Initializable;
import org.clazzes.remoting.InvocationContext;
import org.clazzes.remoting.RemoteInvocation;
import org.clazzes.remoting.RemoteInvocationFailureException;
import org.clazzes.remoting.server.ClientTerminationListener;

/**
 * This ServerInvocationHandler delegates jboss remote calls to a spring bean by
 * intelligently discovering the service interface.
 * 
 * @author wglas
 */
public class ReturnCallbackInvocationHandler implements ReturnCallbackHandler, Initializable, Destroyable, ClientTerminationListener {

	private static final Log log = LogFactory.getLog(ReturnCallbackInvocationHandler.class);
		
	private static class CallbackEntry {
		
		public final UID sessionId;
		public final Class<?> iface;
		public final Object target;
		
		public CallbackEntry(UID sessionId, Class<?> iface, Object target) {
			super();
			this.sessionId = sessionId;
			this.iface = iface;
			this.target = target;
		}
	}
		
	private static class MethodInfo {
		
		public final Method method;
		public final boolean closeMethod;
		
		public MethodInfo(Method method, boolean closeMethod) {
			super();
			this.method = method;
			this.closeMethod = closeMethod;
		}
	}

	private Map<Class<?>,Map<String,MethodInfo>> callbackServiceMethods;
	private Map<Class<?>,List<Method>> callbackCloseMethods;
    private List<Class<?>> callbackInterfaces;
		
	private String closeMethodRegex;
	
	private Map<UID,CallbackEntry> callbackRegistry;
	private Map<UID,Set<UID>> callbacksBySessionId;
	
	private synchronized CallbackEntry getCallbackEntry (UID guid) {
		
		return this.callbackRegistry.get(guid);
	}
		
	/* (non-Javadoc)
     * @see org.clazzes.remoting.Initializable#isInitialized()
     */
    public boolean isInitialized() {
       
        return this.callbackRegistry != null;
    }

    /* (non-Javadoc)
	 * @see org.clazzes.remoting.Initializable#initialize()
	 */
	public void initialize() throws Exception {

		this.callbackServiceMethods = new HashMap<Class<?>, Map<String,MethodInfo>>();
		this.callbackCloseMethods = new HashMap<Class<?>, List<Method>>();
        
		Pattern closePattern =
			this.closeMethodRegex == null ? null : Pattern.compile(this.closeMethodRegex);
		
		for (Class<?> serviceInterface : this.callbackInterfaces) {
		
			Method [] methods = serviceInterface.getMethods();
		
			Map<String, MethodInfo> serviceMethods = new HashMap<String, MethodInfo>();
			List<Method> closeMethods = null;
	        
			for (Method method : methods) {
			
				MethodInfo om = serviceMethods.get(method.getName());
				
				if (om != null)
					throw new InvalidClassException("Method ["+om.method+"] is duplicated in interface ["+serviceInterface.getName()+"] (duplicate is ["+method+"]).");
				
				boolean closeMethod = closePattern == null ? false :
					closePattern.matcher(method.getName()).matches();
				
				if (log.isDebugEnabled()) {
					if (closeMethod)
						log.debug("Adding close method ["+ method +"] to the list of service methods for interface ["+serviceInterface.getName()+"].");
					else
						log.debug("Adding method ["+ method +"] to the list of service methods for interface ["+serviceInterface.getName()+"].");
				}
				
				serviceMethods.put(method.getName(),new MethodInfo(method,closeMethod));
				
				if (closeMethod) {    
				    if (closeMethods == null) closeMethods = new LinkedList<Method>();
				    closeMethods.add(method);
				}
			}
			this.callbackServiceMethods.put(serviceInterface,serviceMethods);
			
			if (closeMethods != null)
			    this.callbackCloseMethods.put(serviceInterface,closeMethods);
		}
		
		this.callbackRegistry = new HashMap<UID,CallbackEntry>();
		this.callbacksBySessionId = new HashMap<UID, Set<UID>>();
	}

	/* (non-Javadoc)
	 * @see javax.security.auth.Destroyable#destroy()
	 */
	public void destroy() {
		
		if (this.callbackRegistry != null) {
			
			if (log.isDebugEnabled())
				log.debug("Found ["+this.callbackRegistry.size()+"] left-over callback objects from ["+this.callbacksBySessionId.size()+"] sessions.");
			
			else if (this.callbackRegistry.size() > 0) {
				
				log.warn("Found ["+this.callbackRegistry.size()+"] left-over callback objects from ["+this.callbacksBySessionId.size()+"] sessions.");
			}
			
			this.callbackRegistry = null;
			this.callbacksBySessionId = null;
		}
	}
	
    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationHandler#invoke(org.clazzes.remoting.InvocationContext, java.lang.Object, java.util.Map)
     */
    public Object invoke(InvocationContext ctx, Object arg,
			Map<String, Object> metadata) throws Throwable {

		if (arg instanceof RemoteInvocation) {
			RemoteInvocation ri = (RemoteInvocation)arg;
			
			if (metadata == null)
				throw new RemoteInvocationFailureException("The invocation request has not request payload.");
				
			UID guid = (UID)metadata.get(RETURN_PROXY_GUID_PROPERTY);
			
			if (guid == null)
				throw new RemoteInvocationFailureException("The invocation request contains no ["+RETURN_PROXY_GUID_PROPERTY+"] property.");

			CallbackEntry entry = this.getCallbackEntry(guid);
			
			if (entry == null)
				throw new RemoteInvocationFailureException("The proxy guid ["+guid+"] in session Id ["+ctx.getClientUID()+"] is unknown or expired.");
			
			MethodInfo m = this.callbackServiceMethods.get(entry.iface).get(ri.getMethodName());
			
			if (m == null)
				throw new RemoteInvocationFailureException("The service Iterface ["+entry.iface.getName()+"] contains no method named ["+ri.getMethodName()+"].");

			if (log.isDebugEnabled())
				log.debug("Calling method ["+ri.getMethodName()+"] on interface ["+entry.iface+"] on proxy ["+guid+"]...");
			
			try {
						
				Object ret = m.method.invoke(entry.target,ri.getArguments());
			
				if (log.isDebugEnabled())
					log.debug("Sucessfully called method ["+ri.getMethodName()+"] on interface ["+entry.iface+"] on proxy ["+guid+"].");
				
				return ret;
			}
			catch (InvocationTargetException e) {
				
				if (log.isDebugEnabled())
					log.debug("Caught exception in method ["+ri.getMethodName()+"] of interface ["+entry.iface+"] on proxy ["+guid+"].",e.getCause());
				throw e.getCause();
				
			} finally {
				
				if (m.closeMethod) {
					
					if (log.isDebugEnabled())
						log.debug("Deregistering callback after close method ["+ri.getMethodName()+"] of interface ["+entry.iface+"] on proxy ["+guid+"].");
					
					this.deregisterCallback(ctx.getClientUID(),guid);
				}
			}
		}
		
		// special case: null argument
		// return number of registered callbacks.
		// This is used for debugging.
		if (arg == null) {
		    return Integer.valueOf(this.countCallbacks(ctx.getClientUID()));
		}
		
		throw new RemoteInvocationFailureException("The invocation parameter ["+arg+"] is not of type RemoteInvocation");
	}

	/* (non-Javadoc)
     * @see org.clazzes.remoting.beans.ReturnCallbackHandler#countCallbacks(java.rmi.server.UID)
     */
    public synchronized int countCallbacks(UID sessionId) {

        return this.callbacksBySessionId.get(sessionId).size();
    }

    /* (non-Javadoc)
	 * @see at.cdes.remoting.server.ReturnCallbackHandler#deregisterAllCallbacks(java.lang.String)
	 */
	public synchronized void deregisterAllCallbacks(UID sessionId) {
	    
        List<CallbackEntry> entries = new LinkedList<CallbackEntry>();
        this.deregisterAllCallbacksInner(sessionId,entries);
        
        for (CallbackEntry e : entries) {
            
            List<Method> closeMethods = this.callbackCloseMethods.get(e.iface);
            
            if (closeMethods != null) {
                
                for (Method closeMethod : closeMethods) {
                 
                    try {
                        
                        if (log.isDebugEnabled())
                            log.info("Calling close method ["+closeMethod+"] upon disconnect of client ["+sessionId+"].");
                            
                        closeMethod.invoke(e.target);
                        
                    } catch (Throwable e1) {
                        
                        log.warn("Caught exception calling close method ["+closeMethod+"] upon disconnect of client ["+sessionId+"]",e1);
                    }
                }
            }
        }
	}
	
	private synchronized void deregisterAllCallbacksInner(UID sessionId, List<CallbackEntry> entries) {
		
		if (this.callbacksBySessionId == null) return;
		
		Set<UID> guids = this.callbacksBySessionId.remove(sessionId);
		
		if (guids == null) return;
		
		if (log.isDebugEnabled()) 
			log.debug("Deregistering ["+guids.size()+"] callbacks for session Id ["+sessionId+"].");

		for (UID guid : guids) {
			
			if (log.isDebugEnabled()) 
				log.debug("callback for session Id ["+sessionId+"] and GUID ["+guid+"] has been deregistered.");
			
			CallbackEntry e = this.callbackRegistry.remove(guid);
			
			if (entries != null)
			    entries.add(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.clazzes.remoting.server.ClientTerminationListener#clientTerminated(java.rmi.server.UID)
	 */
	public void clientTerminated(UID clientUID) {
	    
		this.deregisterAllCallbacks(clientUID);
	}

	/* (non-Javadoc)
	 * @see at.cdes.remoting.server.ReturnCallbackHandler#deregisterCallback(java.lang.String, org.jboss.util.id.GUID)
	 */
	public synchronized void deregisterCallback(UID sessionId, UID guid) {
		
		if (log.isDebugEnabled()) 
			log.debug("Deregistering callback for session Id ["+sessionId+"] and GUID ["+guid+"].");

		CallbackEntry entry = this.callbackRegistry.remove(guid);
		
		if (entry == null) {
			log.warn("deregisterCallback: No callback for session Id ["+sessionId+"] and GUID ["+guid+"] could be found.");	
		}
		else {
			
			if (!sessionId.equals(entry.sessionId)) {
				log.warn("deregisterCallback: callback for GUID ["+guid+"] has been registered for session Id ["+sessionId+"] deregistered in session Id ["+entry.sessionId+"].");	
			}
			
			Set<UID> guids = this.callbacksBySessionId.get(entry.sessionId);
		
			if (guids != null && guids.remove(guid)) {
				
				if (guids.size() == 0)
					this.callbacksBySessionId.remove(entry.sessionId);
			}
			else {
				
				log.warn("deregisterCallback: No callback for GUID ["+guid+"] has not been associated to session Id ["+entry.sessionId+"].");
			}
		}
	}

	/* (non-Javadoc)
	 * @see at.cdes.remoting.server.ReturnCallbackHandler#isCallbackInterface(java.lang.Class)
	 */
	public boolean isCallbackInterface(Class<?> iface) {
		
		return this.callbackServiceMethods.containsKey(iface);
	}

	/* (non-Javadoc)
	 * @see at.cdes.remoting.server.ReturnCallbackHandler#registerCallback(java.lang.String, java.lang.Class, java.lang.Object)
	 */
	public synchronized UID registerCallback(UID sessionId, Class<?> iface, Object target) {
		
		UID guid = new UID();
		
		if (log.isDebugEnabled()) 
			log.debug("Registering callback for session Id ["+sessionId+"] and GUID ["+guid+"], interface ["+iface.getName()+"].");

		this.callbackRegistry.put(guid,new CallbackEntry(sessionId,iface,target));
		
		Set<UID> guids = this.callbacksBySessionId.get(sessionId);
		
		if (guids == null) {
			
			guids = new HashSet<UID>();
			this.callbacksBySessionId.put(sessionId,guids);
		}
		guids.add(guid);
		
		return guid;
	}

	/**
	 * @return the callbackInterfaces
	 */
	public List<Class<?>> getCallbackInterfaces() {
		return this.callbackInterfaces;
	}

	/**
	 * @param callbackInterfaces the callbackInterfaces to set
	 */
	public void setCallbackInterfaces(List<Class<?>> callbackInterfaces) {
		this.callbackInterfaces = callbackInterfaces;
	}

	/**
	 * @return The regular expression, which is used to identify context lose methods.
	 *         A context close method deregisters the callback object and hence marks the end
	 *         of the lifecycle of a return proxy.
	 */
	public String getCloseMethodRegex() {
		return this.closeMethodRegex;
	}

	/**
	 * @param closeMethodRegex the closeMethodRegex to set
	 */
	public void setCloseMethodRegex(String closeMethodRegex) {
		this.closeMethodRegex = closeMethodRegex;
	}

}
