/***********************************************************
 * $Id: Client.java 138 2009-10-01 09:16:19Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 15.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.client;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.URI;
import java.rmi.ConnectException;
import java.rmi.server.UID;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.OneWayHandler;
import org.clazzes.remoting.RemoteInvoker;
import org.clazzes.remoting.cmd.Cmd;
import org.clazzes.remoting.cmd.CmdVisitorSupport;
import org.clazzes.remoting.cmd.ExchangeClientUIDCmd;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.cmd.LoadClassCmd;
import org.clazzes.remoting.helper.InetAddressHelper;
import org.clazzes.remoting.loading.ClassBytes;
import org.clazzes.remoting.marshal.Marshaler;

/**
 * A remote invocation client, which support callbacks over a single TCP/IP.
 * 
 * @author wglas
 */
public class Client implements RemoteInvoker {

    private static final Log log = LogFactory.getLog(Client.class);
    
    private String scheme;
    private InetSocketAddress address;
    private ClientParameters clientParameters;
    private ClassLoader marshalClassLoader;
    private Map <UID,InvocationHandler> callbackHandlers;
    private Map <String,OneWayHandler> broadcastHandlers;
    
    private ExecutorService executorService;
    private boolean ownExecutorService;

    // variables used for connection maintainance.
    private ClientConnectionHandler connectionHandler;
    
    /**
     * Construct a client, which will connect to the given URI.
     * 
     * Please note, that any name resolution required will be performed by
     * {@link Client#connect()}.
     * 
     * The following default parameters will be set:
     * <table>
     *   <tr><td>soTimeout</td><td>60000</td></tr>
     *   <tr><td>tcpNoDelay</td><td>true</td></tr>
     * </table>
     * 
     * @param uri An URI with a supported scheme. Currently, only the <code>tcp</code>
     *           and <code>ssl</code> schemes are supported.
     * @throws IOException If the URI has an unsupported scheme.
     */
    public Client(URI uri) throws IOException {
        
        this.scheme = uri.getScheme();
        
        if ((!"tcp".equals(this.scheme)) && (!"ssl".equals(this.scheme)))
            throw new IOException("Scheme ["+this.scheme+"] is not supported.");
        
        this.address = InetSocketAddress.createUnresolved(uri.getHost(),uri.getPort());
    }
    
    /**
     * @return The client configuration object. If not set via {@link #setClientParameters(ClientParameters)},
     *           a new instance of {@link ClientParameters} or {@link SSLClientParameters} is returned
     *           depending on the scheme of the URI passed to the constructor.
     */
    public ClientParameters getClientParameters() {
        
        if (this.clientParameters == null) {
            
        if ("ssl".equals(this.scheme)) 
            this.clientParameters = new SSLClientParameters();
        else
            this.clientParameters = new ClientParameters();
        }
        
        return this.clientParameters;
    }

    /**
     * @param clientParameters The client parameters to set.
     */
    public void setClientParameters(ClientParameters clientParameters) {
        this.clientParameters = clientParameters;
    }

    /**
     * @return The class loader used for resolving deserialized objects. If not
     *         set via {@link #setMarshalClassLoader(ClassLoader)}, the context class
     *         loader of the current thread is used.
     */
    public ClassLoader getMarshalClassLoader() {
        
        if (this.marshalClassLoader == null)
            this.marshalClassLoader = Thread.currentThread().getContextClassLoader();
        
        return this.marshalClassLoader;
    }

    /**
     * @param classLoader The custom class loader used for resolving deserialized objects.
     */
    public void setMarshalClassLoader(ClassLoader classLoader) {
        this.marshalClassLoader = classLoader;
    }

    /**
     * @return The executor service used to execute the asynchronous reader process.
     *         If no set via {@link #setExecutorService(ExecutorService)}, the result
     *         of {@link Executors#newSingleThreadExecutor()} is returned.
     */
    public ExecutorService getExecutorService() {
        
        if (this.executorService == null) {
            this.executorService = Executors.newSingleThreadExecutor();
            this.ownExecutorService = true;
        }
        
        return this.executorService;
    }

    /**
     * @param executorService The executor service used to execute the asynchronous reader process.
     */
    public void setExecutorService(ExecutorService executorService) {
        this.executorService = executorService;
        this.ownExecutorService = false;
    }

    private static class GetClientUIDVisitor extends CmdVisitorSupport {
        
        UID clientUID;
        
        @Override
        public void visit(ExchangeClientUIDCmd cmd) {
            this.clientUID = cmd.getClientUID();
        }
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#connect()
     */
    public synchronized void connect() throws IOException {
        
        if (this.connectionHandler != null)
            throw new ConnectException("Try to connect an already conencted client.");
        
        Marshaler m = this.getClientParameters().getMarshalerFactory().newMarshaler();

        try {
            Socket socket =
                this.getClientParameters().getSocketFactory().createSocket();
            
            // Setting an soTimeout during connection establishment helps to bail
            // out firewall problems an the like.
            //
            // Specifically, a connection establishment with a SYN packet that is never
            // answered by a FIN or ACK package typically blocks for minutes in an operating
            // system dependent manner.
            //
            // Additionally, if a server is not able to communicate the client UID in a timely
            // fashion, the server may be overloaded or deadlocked. 
            //
            Map<String,Object> meta = this.getClientParameters().tuneSocket(socket);
            
            InetSocketAddress addr =
                InetAddressHelper.resolveSocketAddress(this.address);
            
            if(log.isDebugEnabled())
                log.debug("Connecting to ["+addr+"]...");
            
            socket.connect(addr,this.getClientParameters().getSoTimeout());
        
            m.setup(socket.getOutputStream(),socket.getInputStream(),
                    this.getMarshalClassLoader());
     
            // Fetch the client UID from the server during an initial handshake.
            Object obj = m.readObject();
            
            GetClientUIDVisitor v = new GetClientUIDVisitor();
            
            Cmd.acceptCmd(obj,v);
            
            // The ConnectionHandler is not able to distinguish between a long lasting
            // remote call and a real socket timeout, so we set the socket to blocking
            // operation from here on.
            socket.setSoTimeout(0);
            
            if (this.clientParameters.isAsynchronous()) {
                ThreadedClientConnectionHandler ch = new ThreadedClientConnectionHandler(m,v.clientUID,this,meta);
            
                this.getExecutorService().execute(ch);
                this.connectionHandler = ch;
            }
            else {
                this.connectionHandler = new SimpleClientConnectionHandler(m,v.clientUID,this,meta);  
            }
         
            log.info("Client ["+v.clientUID+"] has been successfully connected to ["+addr+"].");
            
            m = null;
            
        } catch (ClassNotFoundException e) {
            
            log.error("Class resolution error in startup handshake.",e);
            throw new IOException("Class resolution error in startup handshake.");
            
        } finally {
            
            if (m != null) {
                m.close();
                this.connectionHandler = null;
            }
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#deregisterCallback(java.rmi.server.UID)
     */
    public void deregisterCallback(UID handlerUID) {
        
        InvocationHandler handler = null;
        
        if (this.callbackHandlers != null) {
        
            synchronized(this.callbackHandlers) {
                handler = this.callbackHandlers.remove(handlerUID);
            }
        }
        
        if (handler == null)
            log.warn("Trying to deregister callback for non-existent UID ["+handlerUID+"]");
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#getCallback(java.rmi.server.UID)
     */
    public InvocationHandler getCallback(UID handlerUID) {
        
        if (this.callbackHandlers != null) {
            
            synchronized(this.callbackHandlers) {
                return this.callbackHandlers.get(handlerUID);
            }
        }
        
        return null;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#disconnect()
     */
    public synchronized void disconnect() {
        
        if (this.connectionHandler != null) {
        
            UID clientUID = this.connectionHandler.getClientUID();
            
            this.connectionHandler.close();
            this.connectionHandler = null;
            
            log.info("Client ["+clientUID+"] has been successfully disonnected from ["+this.address+"].");

            this.connectionHandler = null;

            if (this.ownExecutorService && this.executorService != null) {
                this.executorService.shutdown();
                this.executorService = null;
                this.ownExecutorService = false;
            }
        }
        else {
            log.warn("Try to diconnect a client from ["+this.address+"] which is already disconnected.");
        }
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#invoke(java.lang.String, java.lang.Object, java.util.Map)
     */
    public Object invoke(String subSystem, Object arg,
            Map<String, Object> metadata) throws InvocationTargetException,
            IOException {
       
        ClientConnectionHandler handler;
          
        // avoid lock over the whole span of a remote call, since
        // then long-lasting call might prevent us from shutting down a client
        // server system.
        synchronized(this) {
        
            if (this.connectionHandler == null)
                throw new IOException("Invocation on an unconnected client.");
            
            handler = this.connectionHandler;
        }
        
        InvocationCmd cmd = new InvocationCmd(subSystem,arg,metadata);

        return handler.send(cmd);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#loadClass(java.lang.String)
     */
    public ClassBytes loadClass(String className) throws IOException {
        
        ClientConnectionHandler handler;
        
        // avoid lock over the whole span of a remote call, since
        // then long-lasting call might prevent us from shutting down a client
        // server system.
        synchronized(this) {
        
            if (this.connectionHandler == null)
                throw new IOException("Try to load a class on an unconnected client.");
            
            handler = this.connectionHandler;
        }
        
        LoadClassCmd cmd = new LoadClassCmd(className);

        return handler.send(cmd);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#isConnected()
     */
    public synchronized boolean isConnected() {
        
        return this.connectionHandler != null;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#registerCallback(org.clazzes.remoting.InvocationHandler)
     */
    public synchronized UID registerCallback(InvocationHandler handler) {
        
        if (this.callbackHandlers == null)
            this.callbackHandlers = new HashMap<UID, InvocationHandler>();
            
        UID uid = new UID();

        synchronized(this.callbackHandlers) {
               
            this.callbackHandlers.put(uid,handler);
        }
        
        return uid; 
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#deregisterBroadcastHandler(java.lang.String)
     */
    public void deregisterBroadcastHandler(String subSystem) {
        
        OneWayHandler handler=null;
        
        if (this.broadcastHandlers != null) {
        
            synchronized(this.broadcastHandlers) {
                handler = this.broadcastHandlers.remove(subSystem);
            }
        }
        
        if (handler == null)
            log.warn("Trying to deregister callback for non-existent subsystem ["+subSystem+"]");

    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#registerBroadcastHandler(java.lang.String, org.clazzes.remoting.OneWayHandler)
     */
    public synchronized void registerBroadcastHandler(String subSystem, OneWayHandler handler)
            throws IOException {
        
        if (this.broadcastHandlers == null)
            this.broadcastHandlers = new HashMap<String, OneWayHandler>();
        
        synchronized(this.broadcastHandlers) {
            this.broadcastHandlers.put(subSystem,handler);
        }
        
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.RemoteInvoker#getBroadcastHandler(java.lang.String)
     */
    public OneWayHandler getBroadcastHandler(String subSystem) {

        if (this.broadcastHandlers != null) {
            
            synchronized(this.broadcastHandlers) {
                return this.broadcastHandlers.get(subSystem);
            }
        }
        return null;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationContext#getClientUID()
     */
    public synchronized UID getClientUID() {
        
        return this.connectionHandler == null ? null : this.connectionHandler.getClientUID();
    }
}
