/***********************************************************
 * $Id: ClassHelper.java 150 2009-10-12 18:52:25Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 01.10.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.helper;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * @author wglas
 *
 */
public abstract class ClassHelper {

    /**
     * @param className
     * @return Whether the given class is an array class, i.e. the
     *         name starts with any number of open brackets followed by
     *         'L' and ends with ';'.
     */
    static public boolean isArrayClass(String className) {
        
        return className.charAt(0) == '[';
    }
    
    /**
     * @param className The name of a class, which may be an array.
     * @return The class name of the array component. If the class is no array,
     *         return <code>className</code>.
     */
    static public String getArrayComponentClass(String className) {
        
        if (!isArrayClass(className)) return className;
        
        int i=1;
        int l = className.length() - 1;
        
        for (;i<l;++i)
            if (className.charAt(i) != '[') break;
   
        switch (className.charAt(i)) {
        
        case 'Z': return boolean.class.getName();
        case 'B': return byte.class.getName();
        case 'C': return char.class.getName();
        case 'S': return short.class.getName();
        case 'I': return int.class.getName();
        case 'J': return long.class.getName();
        case 'D': return double.class.getName();
        case 'F': return float.class.getName();
        
        case 'L':
            if (className.charAt(l) != ';')
                throw new IllegalArgumentException("Array class name ["+className+"] is invalid.");
            
            return className.substring(i+1,l);
            
        default:
            throw new IllegalArgumentException("Array class name ["+className+"] contains an invalid memeber designator.");
        }
    }
    
    /**
     * Return the compiled class from a class loader. 
     * 
     * @param className The name of the class. If this is the name of an array class,
     *                  the compiled class of the array component will be returned.
     * @param cl The class loader to resolve the class resource. If <code>null</code>,
     *           the system class loader will be used.
     * @return The content of the class-file for the given class.
     * @throws IOException Upon read errors.
     * @throws ClassNotFoundException When the .class resource could not be found.
     */
    static public byte[] getCompiledClass(String className, ClassLoader cl) throws IOException, ClassNotFoundException {
 
        if (cl == null) cl = ClassLoader.getSystemClassLoader();
        
        String rn = ClassHelper.getArrayComponentClass(className).replace('.','/')+".class";
        
        InputStream is = cl.getResourceAsStream(rn);
        
        if (is == null) throw new ClassNotFoundException("The resource ["+rn+"] could not be found.");
        
        ByteArrayOutputStream bos = new ByteArrayOutputStream(16384);
        
        byte [] buf = new byte[1024];
        
        int n;
        
        while ((n = is.read(buf)) > 0) {
            
            bos.write(buf,0,n);
        }
        
        return bos.toByteArray();
    }
}
