/***********************************************************
 * $Id: CompactMarshalerFactory.java 126 2009-08-23 15:07:21Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 17.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.TreeMap;
import java.util.TreeSet;

import org.clazzes.remoting.cmd.CallBackCmd;
import org.clazzes.remoting.cmd.ExchangeClientUIDCmd;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.cmd.InvocationExceptionCmd;
import org.clazzes.remoting.cmd.InvocationResultCmd;
import org.clazzes.remoting.cmd.LoadClassCmd;
import org.clazzes.remoting.cmd.ReturnClassBytesCmd;
import org.clazzes.remoting.loading.ClassBytes;
import org.clazzes.remoting.marshal.impl.CompactMarshaler;
import org.clazzes.remoting.marshal.impl.MarshalerClassRegistry;

/**
 * A marshaler factory, that emits a marshaler, which serializes a java
 * class in a low-bandwidth consuming manner. The downside of this implementation
 * is, that the marshaler needs to know the serializable classes it knows a priori.
 * 
 * @author wglas
 */
public class CompactMarshalerFactory extends MarshalerFactorySupport {

    private boolean extendClasses;
    private boolean flattenCollections;
    private MarshalerClassRegistry classRegistry;
    private ObjectResolver objectResolver;
    
    /**
     * Instantiate a CompactMarshalerFactory with the input and output buffer size
     * set to <code>1460</code>, the most common MSS for TCP over ethernet.
     * 
     * <code>extendClasses</code> is set to <code>true</code>, if you change this setting,
     * you might want to register new classes or call <code>setExtendClasses(true)</code>.
     */
    public CompactMarshalerFactory() {
        
        this.classRegistry = new MarshalerClassRegistry();
        this.classRegistry.registerClass(CallBackCmd.class);
        this.classRegistry.registerClass(ExchangeClientUIDCmd.class);
        this.classRegistry.registerClass(InvocationCmd.class);
        this.classRegistry.registerClass(InvocationExceptionCmd.class);
        this.classRegistry.registerClass(InvocationResultCmd.class);
        this.classRegistry.registerClass(LoadClassCmd.class);
        this.classRegistry.registerClass(ReturnClassBytesCmd.class);
        this.classRegistry.registerClass(ClassBytes.class);
        this.classRegistry.registerClass(RuntimeException.class);
        this.classRegistry.registerClass(Exception.class);
        this.classRegistry.registerClass(StackTraceElement.class);
        this.classRegistry.registerClass(Number.class);
        this.classRegistry.registerClass(Date.class);
        this.classRegistry.registerClass(java.sql.Date.class);
        this.classRegistry.registerClass(Timestamp.class);
        this.classRegistry.registerClass(GregorianCalendar.class);
        this.classRegistry.registerClass(TreeMap.class);
        this.classRegistry.registerClass(TreeSet.class);
        this.classRegistry.registerClass(LinkedHashMap.class);
      
        this.extendClasses = true;
        this.flattenCollections = true;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.marshal.MarshalerFactory#newMarshaler()
     */
    @Override
    public Marshaler newMarshaler() {

        return new CompactMarshaler(this.classRegistry,this.extendClasses,
                this.flattenCollections,this.objectResolver,
                this.getOutputBufferSize(),this.getInputBufferSize());
    }

    /**
     * @param c The class to register for marshaling. This is needed, if
     *          {@link #isExtendClasses()} is <code>false</code>.
     */
    public void registerClass(Class<?> c) {
        this.classRegistry.registerClass(c);
    }
    
    /**
     * @return Whether the marshaler dynamically extends the list of marshalable classes.
     */
    public boolean isExtendClasses() {
        return this.extendClasses;
    }

    /**
     * @param extendClasses Whether the marshaler should dynamically extend the list
     *                      of marshalable classes.
     */
    public void setExtendClasses(boolean extendClasses) {
        this.extendClasses = extendClasses;
    }

    /**
     * @return The object resolver, which intercepts objects being read or written.
     */
    public ObjectResolver getObjectResolver() {
        return this.objectResolver;
    }

    /**
     * @param objectResolver The object resolver, which intercepts objects
     *                       being read or written.
     */
    public void setObjectResolver(ObjectResolver objectResolver) {
        this.objectResolver = objectResolver;
    }

    /**
     * @return Whether collection are mapped to specialized serializable form.
     *         If set to <code>true</code>, all collections are deserialized
     *         as {@link ArrayList}, {@link HashSet} and {@link HashMap} respectively.
     */
    public boolean isFlattenCollections() {
        return this.flattenCollections;
    }

    /**
     * @param flattenCollections Whether collection are mapped to specialized
     *                           serializable form.
     */
    public void setFlattenCollections(boolean flattenCollections) {
        this.flattenCollections = flattenCollections;
    }
}
