/***********************************************************
 * $Id: ClassFields.java 150 2009-10-12 18:52:25Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 17.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This class holds the filed marshalers for a given class.
 * 
 * @author wglas
 */
class ClassFields implements ObjectMarshaler {
    
    private static final Log log = LogFactory.getLog(ClassFields.class);
    
    private static final Method OBJECT_CTOR;
    
    static {
        
        try {
            OBJECT_CTOR = ObjectStreamClass.class.getDeclaredMethod("newInstance");
            OBJECT_CTOR.setAccessible(true);
            
        } catch (Throwable e) {
            
            throw new RuntimeException("Cannot find ctor ObjectStreamClass.newInstance()",e);
        }
    }
    
    private static final Comparator<Field> FIELD_COMPARATOR =
        new Comparator<Field>() {

        public int compare(Field a, Field b) {
            
            return a.getName().compareTo(b.getName());
        }
        
    };

    private static final String READ_RESOLVE_METHOD_NAME = "readResolve";
    private static final String WRITE_REPLACE_METHOD_NAME = "writeReplace";
    private static final String READ_OBJECT_METHOD_NAME = "readObject";
    private static final String WRITE_OBJECT_METHOD_NAME = "writeObject";
   
    private static Method fetchResolveMethod(Class<?> clazz, String mn) {

        try {
            Method method = clazz.getDeclaredMethod(mn);
            method.setAccessible(true);
            return method;
            
        } catch (SecurityException e) {
            
            throw new RuntimeException("Cannot fetch method ["+mn+"()] of ["+clazz+"]",e);
        } catch (NoSuchMethodException e) {
   
        }
        return null;
    }
    
    private static void checkForReadWriteObject(Class<?> clazz) {
 
        try {
            clazz.getDeclaredMethod(READ_OBJECT_METHOD_NAME,ObjectInputStream.class);
            log.warn("["+clazz+"] has a method named ["+READ_OBJECT_METHOD_NAME+"()].");
            
         } catch (SecurityException e) {
            
            throw new RuntimeException("Cannot fetch method ["+READ_OBJECT_METHOD_NAME+"()] of ["+clazz+"]",e);
        } catch (NoSuchMethodException e) {
   
        }
      
        try {
            clazz.getDeclaredMethod(WRITE_OBJECT_METHOD_NAME,ObjectOutputStream.class);
            log.warn("["+clazz+"] has a method named ["+WRITE_OBJECT_METHOD_NAME+"()].");

        } catch (SecurityException e) {
            
            throw new RuntimeException("Cannot fetch method ["+WRITE_OBJECT_METHOD_NAME+"()] of ["+clazz+"]",e);
        } catch (NoSuchMethodException e) {
   
        }
    }

    private final List<FieldMarshaler> fieldMarshalers;
    private final ClassFields superClassFields;
    final Method writeReplaceMethod;
    final Method readResolveMethod;

    /**
     * Construct a field description with hand-crafted field marshalers. 
     * @param fieldMarshalers The filed marshalers.
     * @param superClassFields The field description of the super class.
     */
    public ClassFields(List<FieldMarshaler> fieldMarshalers,ClassFields superClassFields) {
        
        this.fieldMarshalers = fieldMarshalers;
        this.superClassFields = superClassFields;
        this.writeReplaceMethod = null;
        this.readResolveMethod = null;
    }
    
    /**
     * Construct a field description by inspecting a class.
     * @param clazz The class to inspect.
     * @param superClassFields The field description of the super class.
     */
    public ClassFields(Class<?> clazz,ClassFields superClassFields) {
        this(clazz,null,superClassFields);
    }
     
    /**
     * Construct a field description by inspecting a class and skipping a list of given fields.
     * @param clazz The class to inspect.
     * @param skipFields The list of fields to skip.
     * @param superClassFields The field description of the super class.
     */
   public ClassFields(Class<?> clazz, Set<String> skipFields,ClassFields superClassFields) {
        super();
            
        ArrayList<Field> ff = new ArrayList<Field>();
       
        Field[] fields = clazz.getDeclaredFields();
            
        for (Field field:fields) {
                
            if (Modifier.isStatic(field.getModifiers()) ||
                    Modifier.isTransient(field.getModifiers()))
                    continue;
            
            if (skipFields != null && skipFields.contains(field.getName()))
                continue;
            
            ff.add(field);
        }

        Collections.sort(ff,FIELD_COMPARATOR);
        
        this.fieldMarshalers = new ArrayList<FieldMarshaler>(ff.size());
        
        for (Field field:ff) {
            
            this.fieldMarshalers.add(FieldMarshaler.newInstance(field));
        }
        this.superClassFields = superClassFields;
        
        Method rrm = fetchResolveMethod(clazz,READ_RESOLVE_METHOD_NAME);
        if (rrm == null && superClassFields != null)
            rrm = superClassFields.getReadResolveMethod();
        this.readResolveMethod = rrm;
        
        Method wrm = fetchResolveMethod(clazz,WRITE_REPLACE_METHOD_NAME);
        if (wrm == null && superClassFields != null)
            wrm = superClassFields.getWriteReplaceMethod();
        this.writeReplaceMethod = wrm;
        
        if (clazz != Enum.class && clazz != Throwable.class)
            checkForReadWriteObject(clazz);
   }

    /**
     * @return the fieldMarshalers
     */
    public List<FieldMarshaler> getFieldMarshalers() {
        return this.fieldMarshalers;
    }

    /**
     * @return the class fields of the super class to be serialized.
     */
    public ClassFields getSuperClassFields() {
        return this.superClassFields;
    }

    /**
     * @return The <code>writeReplace()</code> method of the represented class or null,
     *         if this class has not write replacement.
     */
    public Method getWriteReplaceMethod() {
        return this.writeReplaceMethod;
    }

    /**
     * @return The <code>readResolve()</code> method of the represented class or null,
     *         if this class has not read replacement.
     */
    public Method getReadResolveMethod() {
        return this.readResolveMethod;
    }

    public Object readObject(CompactMarshaler context, long ref, ClassInfo ci) throws IllegalArgumentException, IllegalAccessException, IOException, ClassNotFoundException {
        
        Object o;
        try {
            Class<?> clazz = Class.forName(ci.getClazz(),true,context.getClassLoader());
            o = OBJECT_CTOR.invoke(ObjectStreamClass.lookup(clazz));
            
        } catch (Exception e) {
            
            throw new IllegalArgumentException("Cannot instantiate object of type ["+ci.getClazz()+
                    "] when reading object field.",e); 
        }
        
        context.cacheReadPremature(ref,o);
        
        ClassFields cf = this;
        
        while (cf != null) {
    
            for (FieldMarshaler fm : cf.getFieldMarshalers()) {
            
                if (log.isDebugEnabled())
                    log.debug("Reading field ["+fm.getField()+"]");
            
                fm.readField(context, o);
            }
            cf=cf.getSuperClassFields();
        }
        
        return o;
    }

    public void writeObject(CompactMarshaler context, Object obj) throws IllegalArgumentException, IOException, IllegalAccessException, ClassNotFoundException {
        
        ClassFields cf = this;
        
        while (cf != null) {
            for (FieldMarshaler fm : cf.getFieldMarshalers()) {
                
                if (log.isDebugEnabled())
                    log.debug("Writing field ["+fm.getField()+"]");
                
                fm.writeField(context,obj);
            }
            cf=cf.getSuperClassFields();
        }

    }
}
