/***********************************************************
 * $Id: MarshalerClassRegistry.java 127 2009-08-23 19:27:42Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 20.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.lang.reflect.Field;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.FlatRemoteException;
import org.clazzes.remoting.annotations.OldSerialVersionUID;
import org.clazzes.remoting.annotations.OldSerialVersionUIDs;

/**
 * A registry for known classes to marshal.
 * 
 * @author wglas
 */
public class MarshalerClassRegistry {

    private static final Log log = LogFactory.getLog(MarshalerClassRegistry.class);
    
    private static final ClassInfo FLAT_EXCEPTION_CI = new ClassInfo(FlatRemoteException.class);
    
    private static final Map<Class<?>,ObjectMarshaler> wellKnownClasses;
    
    static {
        
        wellKnownClasses = new HashMap<Class<?>, ObjectMarshaler>();
        wellKnownClasses.put(Date.class,new DateObjectMarshaler());
        wellKnownClasses.put(java.sql.Date.class,new SqlDateObjectMarshaler());
        wellKnownClasses.put(Timestamp.class,new TimestampObjectMarshaler());
        wellKnownClasses.put(GregorianCalendar.class,new GregorianCalendarObjectMarshaler());
        wellKnownClasses.put(Class.class,new ClassObjectMarshaler());
        wellKnownClasses.put(TreeMap.class,new TreeMapObjectMarshaler());
        wellKnownClasses.put(LinkedHashMap.class,new LinkedHashMapObjectMarshaler());
        wellKnownClasses.put(TreeSet.class,new TreeSetObjectMarshaler());
    }
    
    final ConcurrentMap<Class<?>,ClassInfo> classInfos;
    final ConcurrentMap<ClassInfo,ObjectMarshaler> objectMarshalers;
    
    public MarshalerClassRegistry() {
        this.classInfos = new ConcurrentHashMap<Class<?>, ClassInfo>();
        this.objectMarshalers = new ConcurrentHashMap<ClassInfo, ObjectMarshaler>();
    }
    
    private void addOldSVUID(Class<?> c, OldSerialVersionUID osvuid) {

        Set<String> skipFields = new HashSet<String>(osvuid.skipFields().length);
        for (String s : osvuid.skipFields()) {
            skipFields.add(s);
        }
        
        ClassInfo ci = new ClassInfo(c.getName(),osvuid.id());
        
        if (log.isDebugEnabled())
            log.debug("Registering old class info ["+ci+"].");
        
        Class<?> sc = c.getSuperclass();
        ClassFields superCf = null;
            
        if (sc != null) {
     
            ClassInfo superCi;

            if (osvuid.superSerialVersionUID().length() > 0) {
            
                superCi = new ClassInfo(sc.getName(),Long.parseLong(osvuid.superSerialVersionUID()));
            }
            else {
                superCi = new ClassInfo(sc);
            }
            
            superCf = (ClassFields)this.objectMarshalers.get(superCi);
            
            if (superCf == null)
                throw new IllegalArgumentException("Super class ["+sc.getName()+"] of class ["+c.getName()+"] does not contain an @OldSerialVersionUID for UID ["+osvuid.superSerialVersionUID()+"].");
            
        }
        
        ClassFields cf = new ClassFields(c,skipFields,superCf);
        this.objectMarshalers.putIfAbsent(ci,cf);
    }
    
    public ClassInfo registerClass(Class<?> c) {
        
        ObjectMarshaler wkom = wellKnownClasses.get(c);
        
        if (wkom != null)
            return this.registerClass(c,wkom);
        
        if (c.isEnum())
            return this.registerClass(c,EnumObjectMarshaler.INSTANCE);
        
        ClassInfo ci = new ClassInfo(c);
        
        Class<?> sc = c.getSuperclass();
        ClassFields superCf = null;
        
        if (sc != null) {
        
            ClassInfo superCi = this.classInfos.get(sc);
            
            if (superCi == null)
                superCi = this.registerClass(sc);
            
            superCf = (ClassFields)this.objectMarshalers.get(superCi);
        }
        
        this.classInfos.putIfAbsent(c,ci);
        
        if (log.isDebugEnabled())
            log.debug("Registering primary class info ["+ci+"].");

        ClassFields cf = new ClassFields(c,superCf);
        this.objectMarshalers.putIfAbsent(ci,cf);

        OldSerialVersionUID osvuid = c.getAnnotation(OldSerialVersionUID.class);
        
        if (osvuid != null) {    
            this.addOldSVUID(c,osvuid);
        }
        
        OldSerialVersionUIDs osvuids = c.getAnnotation(OldSerialVersionUIDs.class);
        
        if (osvuids != null) {
            
            for (OldSerialVersionUID x:osvuids.value()) {
                this.addOldSVUID(c,x);
            }
        }
        
        if (c == RuntimeException.class) {
            
            // register flattened exception.
            // use a pseudo field marshaler.
            List<FieldMarshaler> marshalers = new ArrayList<FieldMarshaler>();
            FieldMarshaler fm = new ClassNameFieldMarshaler();
            try {
                Field f = FlatRemoteException.class.getDeclaredField("originalClassName");
                fm.setField(f);
            } catch (Exception e) {
                throw new IllegalArgumentException("Field FlatRemoteException.originalClassName could not be found.",e);
            }
            marshalers.add(fm);
            ClassFields fcf = new ClassFields(marshalers,cf);

            this.classInfos.putIfAbsent(FlatRemoteException.class,FLAT_EXCEPTION_CI);
            this.objectMarshalers.putIfAbsent(FLAT_EXCEPTION_CI,fcf);            
        }
        
        return ci;
    }

    private ClassInfo registerClass(Class<?> c, ObjectMarshaler wkom) {        

        ClassInfo ci = new ClassInfo(c);
        
        if (log.isDebugEnabled())
            log.debug("Registering well known class info ["+ci+"].");

        this.classInfos.putIfAbsent(c,ci);
        this.objectMarshalers.putIfAbsent(ci,wkom);
        return ci;
    }
    
    /**
     * Get the default ClassInfo for the given class upon serialization.
     * 
     * @param c The class for which to fetch the default ClassInfo.
     * @param extend Whether the list of registered classes should be extended, if <code>c</code>
     *               is not yet registered.
     * @return The ClassInfo with the actual <code>serialVersionUID</code>.
     * @throws IllegalArgumentException If the class is not yet registered and <code>extend</code>
     *                                  is set to <code>false</code>.
     */
    public ClassInfo getClassInfo(Class<?> c, boolean extend) throws IllegalArgumentException {
        
        ClassInfo ci = this.classInfos.get(c);
        
        if (ci == null) {
        
            if (extend) {
                
                if (log.isDebugEnabled())
                    log.debug("Adding ["+c+"] to the list of marshalable classes during write.");
                ci = this.registerClass(c);
            }
            else {
                
                if (Throwable.class.isAssignableFrom(c))
                    return FLAT_EXCEPTION_CI;
                
                throw new IllegalArgumentException("["+c+"] is not registered for this Marshaler.");
            }
        }
        
        return ci;
    }
    
    /**
     * Get the registered order of fields for the given ClassInfo.
     * 
     * @param ci The class info, typically read from an input stream.
     * @param classLoader If not <code>null</code>, the list of registered classes
     *                    is extended by loading the class with name contained in <code>ci</code>
     *                    from the specified class loader.
     * @return The order list of fields to serialize. 
     * @throws IllegalArgumentException If the given class info could not be found or the given class
     *                    could not be loaded from the given class loader.
     * @throws ClassNotFoundException 
     */
    public ObjectMarshaler getObjectMarshaler(ClassInfo ci, ClassLoader classLoader) throws ClassNotFoundException {

        ObjectMarshaler cf = this.objectMarshalers.get(ci);
        
        if (cf == null && classLoader != null) {
            
            if (log.isDebugEnabled())
                log.debug("Adding ["+ci+"] to the list of marshalable classes during read.");
            this.registerClass(Class.forName(ci.getClazz(),true,classLoader));
            
            cf = this.objectMarshalers.get(ci);
        }

        if (cf == null)
            throw new ClassNotFoundException("["+ci+"] is not registered for this Marshaler.");
        
        return cf;
    }
}
