/***********************************************************
 * $Id: ObjectFieldMarshaler.java 150 2009-10-12 18:52:25Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 17.04.2009
 * 
 ***********************************************************/

package org.clazzes.remoting.marshal.impl;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectStreamClass;
import java.io.UTFDataFormatException;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.marshal.ObjectResolver;


/**
 * This marshaler reads/writes filed of a known object type.
 * 
 * @author wglas
 */
class ObjectFieldMarshaler extends FieldMarshaler {

    private static final Log log = LogFactory.getLog(ObjectFieldMarshaler.class);
    
    private static final Method OBJECT_CTOR;
    
    private static final Map<Class<?>,Integer> primitiveClasses;
    private static final Set<Class<?>> alwaysFlatCollections;

    static {
        
        try {
            OBJECT_CTOR = ObjectStreamClass.class.getDeclaredMethod("newInstance");
            OBJECT_CTOR.setAccessible(true);
            
        } catch (Throwable e) {
            
            throw new RuntimeException("Cannot find ctor ObjectStreamClass.newInstance()",e);
        }
        
        primitiveClasses = new HashMap<Class<?>, Integer>();
        primitiveClasses.put(Character.class, CHAR_STREAM_ID);
        primitiveClasses.put(Boolean.class,   BOOLEAN_STREAM_ID);
        primitiveClasses.put(Byte.class,      BYTE_STREAM_ID);
        primitiveClasses.put(Short.class,     SHORT_STREAM_ID);
        primitiveClasses.put(Integer.class,   INT_STREAM_ID);
        primitiveClasses.put(Long.class,      LONG_STREAM_ID);
        primitiveClasses.put(Float.class,     FLOAT_STREAM_ID);
        primitiveClasses.put(Double.class,    DOUBLE_STREAM_ID);
        primitiveClasses.put(char[].class,    CHAR_ARRAY_STREAM_ID);
        primitiveClasses.put(boolean[].class, BOOLEAN_ARRAY_STREAM_ID);
        primitiveClasses.put(byte[].class,    BYTE_ARRAY_STREAM_ID);
        primitiveClasses.put(short[].class,   SHORT_ARRAY_STREAM_ID);
        primitiveClasses.put(int[].class,     INT_ARRAY_STREAM_ID);
        primitiveClasses.put(long[].class,    LONG_ARRAY_STREAM_ID);
        primitiveClasses.put(float[].class,   FLOAT_ARRAY_STREAM_ID);
        primitiveClasses.put(double[].class,  DOUBLE_ARRAY_STREAM_ID);
        
        alwaysFlatCollections = new HashSet<Class<?>>();
        alwaysFlatCollections.add(HashMap.class);
        alwaysFlatCollections.add(ArrayList.class);
        alwaysFlatCollections.add(HashSet.class);
    }
    
    /**
     * @param field
     */
    public ObjectFieldMarshaler(Field field) {
        super(field);
    }

    private static final void writeString(String s, long ref, DataOutputStream os) throws IOException {
        
        int len = s.length();
        int utfLen = 0;
        
        for (int off = 0; off < len; ++off) {

            char c = s.charAt(off);
            
            if (c <= 0x007F && c != 0) {
                ++utfLen;
            } else if (c > 0x07FF) {
                utfLen += 3;
            } else {
                utfLen += 2;
            }
        }
        
        if (utfLen < 65536) {
            
            os.writeByte(STRING_STREAM_ID);
            os.writeLong(ref);
            os.writeShort(utfLen);
        } else {
            
            os.writeByte(LONG_STRING_STREAM_ID);
            os.writeLong(ref);
            os.writeInt(len);
            os.writeInt(utfLen);
        }
        
        for (int off = 0; off < len; ++off) {

            char c = s.charAt(off);
 
            if (c <= 0x007F && c != 0) {
                os.writeByte((byte) c);
            } else if (c > 0x07FF) {
                os.writeByte((byte) (0xE0 | ((c >> 12) & 0x0F)));
                os.writeByte((byte) (0x80 | ((c >> 6) & 0x3F)));
                os.writeByte((byte) (0x80 | ((c >> 0) & 0x3F)));
            } else {
                os.writeByte((byte) (0xC0 | ((c >> 6) & 0x1F)));
                os.writeByte((byte) (0x80 | ((c >> 0) & 0x3F)));
            }
        }
    }
    
    private static final String readLongString(DataInputStream is) throws IOException {
        
        int len = is.readInt();
        int utfLen = is.readInt();
        
        char[] buf = new char[len];
        
        int c, char2, char3;
        
        int ibyte = 0;
        
        for (int ichar = 0;ichar<len;++ichar) {
            
            ++ibyte;
            if (ibyte > utfLen)
                throw new UTFDataFormatException(
                "malformed input: ["+utfLen+"] UTF bytes did only fill ["+ichar+"] characters out of ["+len+"] nominal characters.");
            c = is.read();
           
            switch (c >> 4) {
            case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
                /* 0xxxxxxx* */
                buf[ichar]=(char)c;
                break;
            case 12: case 13:
                /* 110x xxxx   10xx xxxx*/
                ++ibyte;
                if (ibyte > utfLen)
                    throw new UTFDataFormatException(
                        "malformed input: partial character at end");
                char2 = is.read();
                
                if ((char2 & 0xC0) != 0x80)
                    throw new UTFDataFormatException(
                        "malformed input around byte " + (ibyte-1));
                buf[ichar]=(char)(((c & 0x1F) << 6) | (char2 & 0x3F));
                break;
            case 14:
                /* 1110 xxxx  10xx xxxx  10xx xxxx */
                ibyte += 2;
                if (ibyte > utfLen)
                    throw new UTFDataFormatException(
                        "malformed input: partial character at end");
                char2 = is.read();
                char3 = is.read();
                if (((char2 & 0xC0) != 0x80) || ((char3 & 0xC0) != 0x80))
                    throw new UTFDataFormatException(
                        "malformed input around byte " + (ibyte-2));
                buf[ichar]=(char)(((c & 0x0F) << 12) |
                                  ((char2 & 0x3F) << 6)  |
                                  ((char3 & 0x3F) << 0)   );
                break;
            default:
                /* 10xx xxxx,  1111 xxxx */
                throw new UTFDataFormatException(
                    "malformed input around byte " + (ibyte-1));
            }
        }

        if (ibyte != utfLen)
            throw new UTFDataFormatException(
                    "malformed input: Only ["+ibyte+"] bytes out of ["+utfLen+"] UTF bytes needed to read ["+len+"] nominal characters.");
      
        return new String(buf);
    }
    
    public static Object readObject(CompactMarshaler context) throws IOException, IllegalAccessException, ClassNotFoundException {
        
        int tag = FieldMarshaler.readTag(context);
        
        Object o;
        long ref;
        long ciref;
        ClassInfo ci = null;
        
        switch (tag) {
        
        case NULL_STREAM_ID:
            o=null;
            break;
        
        case OBJECT_REF_STREAM_ID:
            
            ref = context.getDataInputStream().readLong();
            if (log.isDebugEnabled())
                log.debug("Reading reference ["+ref+"].");
            o = context.readCached(ref);
            
            if (o == null) {
                
                o = context.readPrematureCached(ref);
                
                if (o != null) {
                    ObjectResolver or = context.getObjectResolver();
                    // evaluate the user-supplied resolver.
                    if (or != null) {
                        
                        if (log.isDebugEnabled()) {
                            Object o2 = or.resolvePrematureRead(o);
                            if (o2!=o && o != null)
                                log.debug("resolvePrematureRead(): ["+o.getClass().getName()+"].");
                            o=o2;
                        } else {
                            o = or.resolvePrematureRead(o);
                        }
                    }
                }
            }
            
            break;
            
        case STRING_STREAM_ID:
            ref = context.getDataInputStream().readLong();
            o = context.getDataInputStream().readUTF();    
            context.cacheRead(ref, o);
            break;
            
        case LONG_STRING_STREAM_ID:
            ref = context.getDataInputStream().readLong();
            o = readLongString(context.getDataInputStream());
            context.cacheRead(ref, o);
            break;
            
        case CHAR_STREAM_ID:
            o = Character.valueOf(context.getDataInputStream().readChar());
            break;
            
        case BOOLEAN_STREAM_ID:
            o = Boolean.valueOf(context.getDataInputStream().readBoolean());
            break;
            
        case BYTE_STREAM_ID:
            o = Byte.valueOf(context.getDataInputStream().readByte());
            break;
            
        case SHORT_STREAM_ID:
            o = Short.valueOf(context.getDataInputStream().readShort());
            break;
            
        case INT_STREAM_ID:
            o = Integer.valueOf(context.getDataInputStream().readInt());
            break;
            
        case LONG_STREAM_ID:
            o = Long.valueOf(context.getDataInputStream().readLong());
            break;
            
        case FLOAT_STREAM_ID:
            o = Float.valueOf(context.getDataInputStream().readFloat());
            break;
            
        case DOUBLE_STREAM_ID:
            o = Double.valueOf(context.getDataInputStream().readDouble());
            break;
            
       case CHAR_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            char[] a = new char[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readChar();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case BOOLEAN_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            boolean[] a = new boolean[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readBoolean();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case BYTE_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            byte[] a = new byte[context.getDataInputStream().readInt()];
            for (int off = 0; off<a.length;) {
                int n = context.getDataInputStream().read(a, off,a.length-off);
                if (n<=0) throw new EOFException("Try to read a byte array from a clsoed stream.");
                off+=n;
            }
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case SHORT_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            short[] a = new short[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readShort();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case INT_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            int[] a = new int[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readInt();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case LONG_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            long[] a = new long[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readLong();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case FLOAT_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            float[] a = new float[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readFloat();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case DOUBLE_ARRAY_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            double[] a = new double[context.getDataInputStream().readInt()];
            for (int i = 0; i<a.length; ++i)
                a[i] = context.getDataInputStream().readDouble();
            o=a;
            context.cacheRead(ref, o);
        }
        break;
            
        case LIST_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            int sz = context.getDataInputStream().readInt();
            List<Object> list = new ArrayList<Object>(sz);
            
            for (int i=0;i<sz;++i) {
                list.add(ObjectFieldMarshaler.readObject(context));
            }
            o=list;
            context.cacheRead(ref, o);
        }
        break;
        
        case SET_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            int sz = context.getDataInputStream().readInt();
            Set<Object> set = new HashSet<Object>(sz);
            
            for (int i=0;i<sz;++i) {
                set.add(ObjectFieldMarshaler.readObject(context));
            }
            o=set;
            context.cacheRead(ref, o);
        }
        break;
        
        case MAP_STREAM_ID:
        {
            ref = context.getDataInputStream().readLong();
            int sz = context.getDataInputStream().readInt();
            Map<Object,Object> map = new HashMap<Object,Object>(sz);
            
            for (int i=0;i<sz;++i) {
                Object key = ObjectFieldMarshaler.readObject(context);
                Object value = ObjectFieldMarshaler.readObject(context);
                
                map.put(key,value);
            }
            o=map; 
            context.cacheRead(ref, o);
        }
        break;
        
        case OBJECT_ARRAY_STREAM_ID:
            ci = new ClassInfo(context.getDataInputStream().readUTF(),context.getDataInputStream().readLong());
            // fall through

        case OBJECT_ARRAY_CIREF_STREAM_ID:

            ciref = context.getDataInputStream().readLong();
            
            if (ci == null) {
                ci = context.readClassInfoCached(ciref);
            } else {
                context.cacheClassInfoRead(ciref,ci);
            }
            
            ref = context.getDataInputStream().readLong();

            int sz = context.getDataInputStream().readInt();
            o = Array.newInstance(Class.forName(ci.getClazz(),true,context.getClassLoader()),sz);
            
            for (int i = 0; i<sz; ++i)
               Array.set(o,i,readObject(context));
            
            context.cacheRead(ref, o);
            break;
            
        case OBJECT_STREAM_ID:
            ci = new ClassInfo(context.getDataInputStream().readUTF(),context.getDataInputStream().readLong());
            // fall through
            
        case OBJECT_CIREF_STREAM_ID:

            ciref = context.getDataInputStream().readLong();
            
            if (ci == null) {
                ci = context.readClassInfoCached(ciref);
            } else {
                context.cacheClassInfoRead(ciref,ci);
            }
            
            ref = context.getDataInputStream().readLong();

            ObjectMarshaler om = context.getObjectMarshaler(ci);
            
            // ctor, context.cacheRead(ref, o) and field reading.
            o = om.readObject(context,ref,ci);
             
            Method readResolve = om.getReadResolveMethod();
            
            while (readResolve != null) {
                
                Class<?> oclazz = o.getClass();
                
                try {
                    if (log.isDebugEnabled())
                        log.debug("readResolve(): ["+o.getClass().getName()+"].");

                    o = readResolve.invoke(o);
                    
                } catch (Exception e) {
                    throw new IllegalArgumentException("Cannot call readResolve on object of type ["+ci.getClazz()+
                            "] when reading object field.",e); 
                }
                
                // if we are not configured for auto-extension,
                // context.getClassInfo() might throw an exception, if the
                // resolved class is not known. So disable recursive
                // read resolution for the moment.
                if (!context.isExtendClasses()) break;
                // avoid NPE and endless recursion.
                if (o==null || o.getClass() == oclazz) break;
                ci = context.getClassInfo(o.getClass());
                if (ci == null) break;
                om = context.getObjectMarshaler(ci);
                readResolve = om.getReadResolveMethod();
            }
                        
            ObjectResolver or = context.getObjectResolver();
            // evaluate the user-supplied resolver.
            if (or != null) {
                
                if (log.isDebugEnabled()) {
                    Object o2 = or.resolveRead(o);
                    if (o2!=o && o != null)
                        log.debug("resolveRead(): ["+o.getClass().getName()+"].");
                    o=o2;
                } else {
                    o = or.resolveRead(o);
                }
            }
      
            // replace the cached reference with the resolved object.
            // This mimics the behaviour of java.io.ObjectInputStream.
            context.cacheReadResolved(ref, o);
            break;
            
        default:
            throw new IllegalArgumentException("Invalid tag ["+tag+
                    "] found when reading object field."); 
        }
     
        if (log.isDebugEnabled()) {
            if (o == null)
                log.debug("Read [null].");
            else
                log.debug("Read ["+o.getClass().getName()+"].");
        }
        
        return o;
    }
    
    @Override
    public void readField(CompactMarshaler context, Object obj) throws IllegalArgumentException, IllegalAccessException, IOException, ClassNotFoundException {
        
        Object o = readObject(context);
        this.field.set(obj,o);
    }

    @Override
    public void writeField(CompactMarshaler context, Object obj) throws IOException, IllegalArgumentException, IllegalAccessException, ClassNotFoundException {
        
        Object o = this.field.get(obj);
        writeObject(context,o);
    }
    
    protected static boolean writeFlatCollection(CompactMarshaler context, Object o) throws IOException, IllegalArgumentException, IllegalAccessException, ClassNotFoundException {
        
        if (o instanceof Collection<?>) {
            
            long nref = context.cacheWrite(o);
    
            if (o instanceof List<?>) {
        
                List<?> list = (List<?>) o;
                context.getDataOutputStream().write(LIST_STREAM_ID);
                context.getDataOutputStream().writeLong(nref);
                context.getDataOutputStream().writeInt(list.size());
        
                for (Object x:list) {
                    ObjectFieldMarshaler.writeObject(context,x);
                }
                return true;
            }
            else if (o instanceof Set<?>) {
        
                Set<?> set = (Set<?>) o;
                context.getDataOutputStream().write(SET_STREAM_ID);
                context.getDataOutputStream().writeLong(nref);
                context.getDataOutputStream().writeInt(set.size());
        
                for (Object x:set) {
                    ObjectFieldMarshaler.writeObject(context,x);
                }
                return true;
            }
            else
                throw new IllegalArgumentException("Unsupported collection ["+o.getClass()+
                "] found when writing an object."); 
        }
        else if (o instanceof Map<?,?>) {
    
            long nref = context.cacheWrite(o);
    
            Map<?,?> map = (Map<?, ?>) o;
            context.getDataOutputStream().write(MAP_STREAM_ID);
            context.getDataOutputStream().writeLong(nref);
            context.getDataOutputStream().writeInt(map.size());
            
            for (Map.Entry<?,?> entry:map.entrySet()) {
        
                ObjectFieldMarshaler.writeObject(context,entry.getKey());
                ObjectFieldMarshaler.writeObject(context,entry.getValue());
            }
            return true;
        }
        return false;
    }
    
    public static void writeObject(CompactMarshaler context, Object o) throws IOException, IllegalArgumentException, IllegalAccessException, ClassNotFoundException {
        
        
        if (o == null) {
            
            if (log.isDebugEnabled())
                log.debug("Writing [null].");

            context.getDataOutputStream().write(NULL_STREAM_ID);
            return;
        }

        if (log.isDebugEnabled())
            log.debug("Writing ["+o.getClass().getName()+"].");
        
        Long ref = context.writeCached(o);
        
        if (ref == null) {
            
            // Is this a primitive class with a special serialization handling?
            Integer ptag = primitiveClasses.get(o.getClass());
            
            if (ptag != null) {
                
                context.getDataOutputStream().write(ptag.intValue());
                
                switch (ptag.intValue()) {
                
                case CHAR_STREAM_ID:
                    context.getDataOutputStream().writeChar(((Character)o).charValue());
                    break;
                    
                case BOOLEAN_STREAM_ID:
                    context.getDataOutputStream().writeBoolean(((Boolean)o).booleanValue());
                    break;
                    
                case BYTE_STREAM_ID:
                    context.getDataOutputStream().writeByte(((Byte)o).byteValue());
                    break;
                    
                case SHORT_STREAM_ID:
                    context.getDataOutputStream().writeShort(((Short)o).shortValue());
                    break;
                    
                case INT_STREAM_ID:
                    context.getDataOutputStream().writeInt(((Integer)o).intValue());
                    break;
                
                case LONG_STREAM_ID:
                    context.getDataOutputStream().writeLong(((Long)o).longValue());
                    break;
                    
                case FLOAT_STREAM_ID:
                    context.getDataOutputStream().writeFloat(((Float)o).floatValue());
                    break;
                    
                case DOUBLE_STREAM_ID:
                    context.getDataOutputStream().writeDouble(((Double)o).doubleValue());
                    break;
                    
                case CHAR_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((char[])o).length);
                    for (char c:(char[])o)
                        context.getDataOutputStream().writeChar(c);
                    break;
                    
                case BOOLEAN_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((boolean[])o).length);
                    for (boolean c:(boolean[])o)
                        context.getDataOutputStream().writeBoolean(c);
                    break;
                    
                case BYTE_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((byte[])o).length);
                    for (byte c:(byte[])o)
                        context.getDataOutputStream().writeByte(c);
                    break;
                    
                case SHORT_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((short[])o).length);
                    for (short c:(short[])o)
                        context.getDataOutputStream().writeShort(c);
                    break;
                   
                case INT_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((int[])o).length);
                    for (int c:(int[])o)
                        context.getDataOutputStream().writeInt(c);
                    break;
                    
                case LONG_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((long[])o).length);
                    for (long c:(long[])o)
                        context.getDataOutputStream().writeLong(c);
                    break;
                    
                case FLOAT_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((float[])o).length);
                    for (float c:(float[])o)
                        context.getDataOutputStream().writeFloat(c);
                    break;
                    
                case DOUBLE_ARRAY_STREAM_ID:
                    context.getDataOutputStream().writeLong(context.cacheWrite(o));
                    context.getDataOutputStream().writeInt(((double[])o).length);
                    for (double c:(double[])o)
                        context.getDataOutputStream().writeDouble(c);
                    break;
                }
                
            } else if (o instanceof String) {
                
                writeString((String)o,context.cacheWrite(o),context.getDataOutputStream());
                
            } else if (o.getClass().isArray()) { 
                
                long nref = context.cacheWrite(o);
                Class<?> cc = o.getClass().getComponentType();
                ClassInfo ci = context.getClassInfo(cc);
                Long ciref = context.writeClassInfoCached(ci);

                if (ciref == null) {
                    
                    context.getDataOutputStream().write(OBJECT_ARRAY_STREAM_ID);
                    context.getDataOutputStream().writeUTF(ci.getClazz());
                    context.getDataOutputStream().writeLong(ci.getSerialVersionUID());
                    context.getDataOutputStream().writeLong(context.cacheClassInfoWrite(ci));
                    
                } else {
                    
                    context.getDataOutputStream().write(OBJECT_ARRAY_CIREF_STREAM_ID);
                    context.getDataOutputStream().writeLong(ciref.longValue());
                }

                context.getDataOutputStream().writeLong(nref);
                
                int l = Array.getLength(o);
                context.getDataOutputStream().writeInt(l);
                for (int i=0;i<l;++i)
                    writeObject(context,Array.get(o,i));

            } else {
                
                // some collections implement
                //
                // writeObject(ObjectInputStream) and
                // readObject(ObjectOutputStream)
                //
                // and hence need to be serialized in a specialized way by all means.
                //
                if (context.isFlattenCollections() ||
                        alwaysFlatCollections.contains(o.getClass())) {
                    if (writeFlatCollection(context,o))
                        return;
                }
                Class <?> oclazz = o.getClass();
                ClassInfo ci = context.getClassInfo(oclazz);
                ObjectMarshaler om = context.getObjectMarshaler(ci);
                
                Object oout;
                
                Method writeReplace = om.getWriteReplaceMethod();
                
                if (writeReplace != null) {
                        
                    try {
                        oout = writeReplace.invoke(o);
                    } catch (InvocationTargetException e) {
                        throw new IllegalArgumentException("Error calling writeReplace() on object of type ["+o.getClass()+
                                "].",e); 
                    }
                    
                    if (log.isDebugEnabled())
                        log.debug("writeReplace(): ["+oout.getClass().getName()+"].");
                    
                }
                else
                    oout = o;
                    
                ObjectResolver or = context.getObjectResolver();
                    
                if (or != null) {
                    
                    if (log.isDebugEnabled()) {
                    
                        Object o2 =  or.replaceWrite(oout);
                        
                        if (o2 != oout) {
                            log.debug("replaceWrite(): ["+o2.getClass().getName()+"].");
                            oout = o2;
                        }

                    } else {
                        oout = or.replaceWrite(oout);
                    }
                }

                if (oout != o) {
                    
                    // is the replacement cached ?
                    ref = context.writeCached(oout);
                    
                    if (ref != null) {
                        
                        
                        if (log.isDebugEnabled())
                            log.debug("Writing resolved reference ["+ref+"].");
                        
                        context.getDataOutputStream().write(OBJECT_REF_STREAM_ID);
                        context.getDataOutputStream().writeLong(ref.longValue());
                        return;
                    }
                }
                
                if (oout != o && oout.getClass() != oclazz) {
                        
                    // did the write replacement process return something which
                    // has to be serialized with a different tag ?
                    if (primitiveClasses.containsKey(oout.getClass()) ||
                            oout.getClass().isArray() ||
                            (context.isFlattenCollections() &&
                                    (oout instanceof Collection<?> || oout instanceof Map<?, ?>)) ||
                             (!context.isFlattenCollections() &&
                                     alwaysFlatCollections.contains(oout.getClass())))
                    {
                        writeObject(context, oout);
                        return;
                    }
                    
                    ci = context.getClassInfo(oout.getClass());
                    om = context.getObjectMarshaler(ci);
                }
                
                // cache the resolved instance here.
                long nref = context.cacheWrite(oout);
                Long ciref = context.writeClassInfoCached(ci);
            
                if (ciref == null) {
                    
                    context.getDataOutputStream().write(OBJECT_STREAM_ID);
                    context.getDataOutputStream().writeUTF(ci.getClazz());
                    context.getDataOutputStream().writeLong(ci.getSerialVersionUID());
                    context.getDataOutputStream().writeLong(context.cacheClassInfoWrite(ci));
                    
                } else {
                    
                    context.getDataOutputStream().write(OBJECT_CIREF_STREAM_ID);
                    context.getDataOutputStream().writeLong(ciref.longValue());
                }
            
                context.getDataOutputStream().writeLong(nref);
                
                om.writeObject(context,oout);
            }
            
        } else {
            
            if (log.isDebugEnabled())
                log.debug("Writing object reference ["+ref+"].");
            
            context.getDataOutputStream().write(OBJECT_REF_STREAM_ID);
            context.getDataOutputStream().writeLong(ref.longValue());
        }
    }

}
