/***********************************************************
 * $Id: Server.java 138 2009-10-01 09:16:19Z wglas $
 * 
 * remoting framework of the clazzes project.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 16.10.2008
 * 
 ***********************************************************/

package org.clazzes.remoting.server;

import java.io.Closeable;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URI;
import java.rmi.server.UID;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.clazzes.remoting.Broadcastable;
import org.clazzes.remoting.InvocationHandler;
import org.clazzes.remoting.InvocationTarget;
import org.clazzes.remoting.cmd.InvocationCmd;
import org.clazzes.remoting.helper.InetAddressHelper;
import org.clazzes.remoting.marshal.Marshaler;

/**
 * A remoting server, which listens on a TCP/IP port.
 * 
 * @author wglas
 */
public class Server implements InvocationTarget {

    private static final Log log = LogFactory.getLog(Server.class);
    
    private String scheme;
    private InetSocketAddress address;
    private int maxClients;
    private ClassLoader marshalClassLoader;
    private ExecutorService executorService;
    private boolean ownExecutorService;
    private ServerParameters serverParameters;
    
    private ClientTerminationListener clientTerminationListener;

    private Map<String,InvocationHandler> invocationHandlers;
    
    private ListenHandler listenHandler;

    private class ListenHandler implements Runnable, Broadcastable, ServerRegistry {
    
        private final ServerSocket serverSocket;
        private final Map <UID,ServerConnectionHandler> clients;
        
        /**
         * @param serverSocket
         */
        public ListenHandler(ServerSocket serverSocket) {
            super();
            this.serverSocket = serverSocket;
            this.clients = new HashMap<UID,ServerConnectionHandler>(Server.this.maxClients);
        }

        private synchronized boolean isClosed() {
            return this.serverSocket.isClosed();
        }
        
        public void run() {
            
            try {
                while (!this.isClosed()) {
            
                    Socket s = null;
                    
                    Marshaler marshaler = null;
                    
                    UID clientUID = new UID();
                        
                    try {
                        
                        s = this.serverSocket.accept();
                        Map<String,Object> meta = Server.this.getServerParameters().tuneSocket(s);
     
                        if (!this.clientsAvailable()) {
                            s.close();
                            continue;
                        }
                        
                        marshaler =
                            Server.this.getServerParameters().getMarshalerFactory().newMarshaler();
                        
                        marshaler.setup(s.getOutputStream(),s.getInputStream(),Server.this.getMarshalClassLoader());
                        
                        ServerConnectionHandler handler =
                            new ServerConnectionHandler(this,clientUID,marshaler,meta);
                    
                        synchronized(this) {
                            this.clients.put(clientUID,handler);
                        }
                    
                        Server.this.getExecutorService().execute(handler);
                        log.info("Client ["+clientUID+"] from ["+s.getRemoteSocketAddress()+"] has connected.");
                        
                    } catch (Throwable e) {
                        
                        // post marshaler setup exception.
                        if (marshaler != null) {
                             
                            synchronized(this) {
                                this.clients.remove(clientUID);
                            }
                            marshaler.close();
                        }
                        else if (s != null) {
                            
                            // pre-marshaler setup, post-accept() exception.
                            try {
                                s.close();
                            } catch (IOException e2) {
                                log.warn("I/O error during emergency close of accepted socket.",e2);
                            }
                        }
                        else {
                            // socket is null, this is an error from accept().    
                            if (this.isClosed()) {
                                
                                if (log.isTraceEnabled())
                                    log.trace("Socket close exception in listening loop.",e);
                                
                                break;
                             }
                        }
                        
                        log.error("Exception caught while starting handler for client ["+clientUID+"].",e);
                    }   
                }
                
                log.info("Listening loop finished upon socket close."); 
                
            } catch (Throwable e) {
                
                log.error("Exception caught in listening loop.",e);

            } finally {
                
                try {
                    synchronized(this) {
                        if (!this.serverSocket.isClosed()) {
                            
                            log.warn("emergency close of listening socket upon listening loop exit.");
                            this.serverSocket.close();
                        }
                    }
                } catch (IOException e2) {
                    log.warn("I/O error during emergency close of listening socket.",e2);
                }
            }
        }

        public synchronized boolean clientsAvailable() throws InterruptedException {
        
            if (this.clients.size() < Server.this.maxClients) return true;
            
            log.warn("Maximal number of clients ["+Server.this.maxClients+"] has been reached, Terminating client.");
            
            return false;
        }
        
        public synchronized void clientTerminated(UID clientUID) {
           
            Closeable clientHandler = this.clients.remove(clientUID);
            
            if (clientHandler == null)
                log.error("Client ["+clientUID+"] pledges to be finished, but has not been registered before.");
            else {
                
                if (Server.this.clientTerminationListener != null) {
                    try {
                        
                        if (log.isDebugEnabled())
                            log.debug("Calling user-specified termination listener for client ["+clientUID+"]...");
                            
                        Server.this.clientTerminationListener.clientTerminated(clientUID);

                        if (log.isDebugEnabled())
                            log.debug("Successfully called user-specified termination listener for client ["+clientUID+"].");
                        
                    } catch (Throwable e) {
                        
                        log.warn("RuntimeException caught calling the user-specified termination handler for client ["+clientUID+"].",e);
                    }
                }
                
                log.info("Client ["+clientUID+"] has disconnected.");
             }
        }
        
        /* (non-Javadoc)
         * @see org.clazzes.remoting.Broadcastable#broadcast(java.lang.String, java.lang.Object, java.util.Map)
         */
        public synchronized void broadcast(String subSystem, Object arg,
                Map<String, Object> metadata) {
            
            InvocationCmd cmd = new InvocationCmd(subSystem,arg,metadata);
            
            for (Entry<UID,ServerConnectionHandler> clientEntry : this.clients.entrySet()) {
                try {
                    
                    if (log.isDebugEnabled())
                        log.debug("Sending broadcast to subsystem ["+subSystem+"] of client ["+clientEntry.getKey()+"].");
                    
                    clientEntry.getValue().broadcast(cmd);
                    
                } catch (IOException e) {
                    log.warn("I/O error sending broadcast to subsystem ["+subSystem+"] of client ["+clientEntry.getKey()+"].",e);
                }
            }       
        }

        public synchronized void close() {
            log.info("Closing ["+this.clients.size()+"] remaining clients.");
            
            for (Entry<UID,ServerConnectionHandler> clientEntry : this.clients.entrySet()) {
                try {
                    clientEntry.getValue().close();
                } catch (IOException e) {
                    log.warn("I/O error closing client ["+clientEntry.getKey()+"].",e);
                }
            }
                        
            try {
                if (!this.serverSocket.isClosed())
                    this.serverSocket.close();
            } catch (IOException e) {
                log.warn("I/O error closing listening socket.",e);
            }
        }
        
        public InvocationHandler getInvocationHandler(String subSystem) {
            synchronized (Server.this) {
                return Server.this.invocationHandlers.get(subSystem);
            }
        }

        public ClassLoader getMarshalClassLoader() {
            
            return Server.this.getMarshalClassLoader();
        }

    }
    
    /**
     * Construct a server, which will listen to the given URI.
     * 
     * Please note, that any name resolution required will be performed by
     * {@link Server#listen()}.
     * 
     * The following default parameters will be set:
     * <table>
     *   <tr><td>maxClients</td><td>32</td></tr>
     * </table>
     * 
     * @param uri An URI with a supported scheme. Currently, only the <code>tcp</code>
     *            scheme is supported.
     * @throws IOException If the URI has an unsupported scheme.
     */
    public Server(URI uri) throws IOException {
        
        this.scheme = uri.getScheme();
        
        if ((!"tcp".equals(this.scheme)) && (!"ssl".equals(this.scheme)))
             throw new IOException("Scheme ["+this.scheme+"] is not supported.");
        
        this.address = InetSocketAddress.createUnresolved(uri.getHost(),uri.getPort());
        
        this.maxClients = 32;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#broadcast(java.lang.String, java.lang.Object, java.util.Map)
     */
    public void broadcast(String subSystem, Object arg,
            Map<String, Object> metadata) {
        
        
    }

    /**
     * @return The class loader used for resolving deserialized objects. If not
     *         set via {@link #setMarshalClassLoader(ClassLoader)}, the context class
     *         loader of the current thread is used.
     */
    public ClassLoader getMarshalClassLoader() {
        
        if (this.marshalClassLoader == null)
            this.marshalClassLoader = Thread.currentThread().getContextClassLoader();
        
        return this.marshalClassLoader;
    }

    /**
     * @param classLoader The custom class loader used for resolving deserialized objects.
     */
    public void setMarshalClassLoader(ClassLoader classLoader) {
        this.marshalClassLoader = classLoader;
    }

    /**
     * @return The executor service used to execute the asynchronous reader process.
     *         If no set via {@link #setExecutorService(ExecutorService)}, the result
     *         of {@link Executors#newCachedThreadPool()} is returned.
     */
    public ExecutorService getExecutorService() {
        
        if (this.executorService == null) {
            this.executorService = Executors.newCachedThreadPool();
            this.ownExecutorService = true;
        }
        
        return this.executorService;
    }

    /**
     * @param executorService The executor service used to execute the asynchronous reader process.
     */
    public void setExecutorService(ExecutorService executorService) {
        this.executorService = executorService;
        this.ownExecutorService= false;
    }

    /**
     * @return The user-specified termination listener.
     */
    public ClientTerminationListener getClientTerminationListener() {
        return this.clientTerminationListener;
    }

    /**
     * @param clientTerminationListener A user-specified listener, which is used
     *            to inform about the termination of a client.
     */
    public void setClientTerminationListener(
            ClientTerminationListener clientTerminationListener) {
        this.clientTerminationListener = clientTerminationListener;
    }

    /**
     * @return The maximal number of clients used to connect to this server.
     */
    public int getMaxClients() {
        return this.maxClients;
    }

    /**
     * @param maxClients The maximal number of clients to set.
     */
    public void setMaxClients(int maxClients) {
        this.maxClients = maxClients;
    }

    /**
     * @return The server parameters used. If no set via {@link #setServerParameters(ServerParameters)},
     *         a new instance of {@link ServerParameters} or {@link SSLServerParameters} is returned
     *          depending on the scheme of the URI passed to the constructor.
     */
    public ServerParameters getServerParameters() {
        
        if (this.serverParameters == null) {
        
            if ("ssl".equals(this.scheme))
                this.serverParameters = new SSLServerParameters();
            else
                this.serverParameters = new ServerParameters();
        
        }
        
        return this.serverParameters;
    }

    /**
     * @param serverParameters The server parameters to set.
     */
    public void setServerParameters(ServerParameters serverParameters) {
        this.serverParameters = serverParameters;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#listen()
     */
    public synchronized void listen() throws IOException {
        
        if (this.listenHandler != null)
            throw new IOException("Try to listen twice on a server.");
        
        ServerSocket serverSocket =
            this.getServerParameters().getServerSocketFactory().createServerSocket();
        
        this.getServerParameters().tuneServerSocket(serverSocket);
        
        // do the name resolution now.
        InetSocketAddress addr =
            InetAddressHelper.resolveSocketAddress(this.address);
        
        serverSocket.bind(addr,this.getServerParameters().getBacklog());
        
        this.listenHandler = new ListenHandler(serverSocket);
        
        this.getExecutorService().execute(this.listenHandler);
        
        log.info("Remote server ["+addr+"] has been successfully started.");
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#deregisterInvocationHandler(java.lang.String)
     */
    public synchronized void deregisterInvocationHandler(String subSystem) {
        
        InvocationHandler ih = this.invocationHandlers.remove(subSystem);
        
        if (ih == null)
            log.warn("Try to remove invocation handler for subsystem ["+subSystem+"], which apparently has been removed before.");
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#registerInvocationHandler(java.lang.String, org.clazzes.remoting.InvocationHandler)
     */
    public synchronized void registerInvocationHandler(String subSystem,
            InvocationHandler handler) {
        
        if (this.invocationHandlers == null)
            this.invocationHandlers = new HashMap<String, InvocationHandler>();
        
        this.invocationHandlers.put(subSystem,handler);
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#setInvocationHandlers(java.util.Map)
     */
    public synchronized void setInvocationHandlers(Map<String, InvocationHandler> handlers) {
        
        this.invocationHandlers = handlers;
    }

    /* (non-Javadoc)
     * @see org.clazzes.remoting.InvocationTarget#shutdown()
     */
    public synchronized void shutdown() {
        
        if (this.listenHandler != null)
            this.listenHandler.close();
            
        this.listenHandler = null;
        
        if (this.executorService != null && this.ownExecutorService) {
            this.executorService.shutdown();
            this.executorService = null;
            this.ownExecutorService = false;
        }
            
        log.info("Remote server ["+this.address+"] has been shut down.");
    }
}
